SystemOrganization addCategory: #'Scripting-Kernel'!

!DelayWaitTimeout methodsFor: '*Scripting-override' stamp: 'ar 4/1/2009 15:56'!
wait
	"Wait until either the semaphore is signaled or the delay times out"
	[self schedule.
	"Move the Tweak portion out of Semaphore>>wait, otherwise we might 
	be interrupted before primitiveWait and be left suspended in limbo."
	(process := Processor activeProcess) isSystemProcess
		ifFalse:[process beginWaitOn: delaySemaphore].
	"It is critical that the following has no suspension point so that
	the test and the wait primitive are atomic. In addition, if the delay
	is no longer being waited on while entering the way we know that it 
	is expired because the delay has already fired."
	beingWaitedOn 
		ifTrue:[delaySemaphore primitiveWait]
		ifFalse:[expired := true]] ensure:[
			process isSystemProcess ifFalse:[process endWaitOn: self].
			self unschedule.
	].
	^self isExpired
! !

Link subclass: #ScriptEvent
	instanceVariableNames: 'sender selector arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

ScriptEvent subclass: #ScriptChangeEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptChangeEvent methodsFor: 'testing' stamp: 'ar 1/8/2008 17:47'!
coalesceEvents
	"Answer true if events like the receiver should be coalesced or not"
	^true
! !

!ScriptChangeEvent methodsFor: 'accessing' stamp: 'ar 4/21/2005 21:58'!
messageArguments
	"Answer the arguments to be provided for an 'inline' message dispatch."
	^Array 
		with: (self arguments at: 1) "the current value"
		with: (self firstEvent arguments at: 2) "the oldest value"! !

!ScriptChangeEvent methodsFor: 'accessing' stamp: 'ar 4/21/2005 21:59'!
value
	^self arguments at: 1! !

ScriptEvent subclass: #ScriptCollectionEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptCollectionEvent methodsFor: 'testing' stamp: 'ar 1/9/2008 17:53'!
coalesceEvents
	"Answer true if events like the receiver should be coalesced or not"
	^true
! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:33'!
arguments
	^arguments ifNil:[arguments := #()]! !

!ScriptEvent methodsFor: 'testing' stamp: 'ar 1/8/2008 17:47'!
coalesceEvents
	"Answer true if events like the receiver should be coalesced or not"
	^false
! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 1/10/2008 17:23'!
droppedEvent
	"Return the prior (e.g., dropped) event that occured before the receiver"
	^self priorEvent! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 1/10/2008 17:44'!
droppedEventStreamOfSize: sizeNeeded
	self priorEvent ifNil:[
		^(WriteStream on: (Array new: sizeNeeded+1))
			nextPut: self;
			yourself
	] ifNotNil:[
		^(self priorEvent droppedEventStreamOfSize: sizeNeeded+1) 
			nextPut: self;
			yourself.
	].! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 9/8/2004 16:19'!
firstEvent
	"Answer the first (oldest) event in the chain of events"
	| evt |
	evt := self.
	[evt priorEvent == nil] whileFalse:[evt := evt priorEvent].
	^evt! !

!ScriptEvent methodsFor: 'experimental' stamp: 'ar 7/20/2001 21:19'!
handler
	"Return the handler scope that received the event"
	^Processor activeProcess eventHandlerFor: self! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 4/21/2005 21:57'!
messageArguments
	"Answer the arguments to be provided for an 'inline' message dispatch."
	^arguments ifNil:[#()]! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 1/10/2008 17:45'!
priorEvent
	"Return the prior (e.g., dropped) event that occured before the receiver"
	^nextLink! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 1/10/2008 17:23'!
priorEvent: anEvent
	nextLink := anEvent! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:33'!
selector
	^selector! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:33'!
sender
	^sender! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 7/20/2001 18:35'!
setArguments: args
	arguments := args.! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 2/9/2004 18:06'!
setSender: newSender
	sender := newSender.! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 7/14/2001 16:33'!
setSender: aSender selector: aSelector
	sender := aSender.
	selector := aSelector.! !

!ScriptEvent methodsFor: 'private' stamp: 'ar 7/14/2001 16:33'!
setSender: aSender selector: aSelector arguments: args
	sender := aSender.
	selector := aSelector.
	arguments := args.! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 7/20/2001 20:55'!
withDroppedEvents
	"Return an array of events that contain the receiver and all dropped events from the oldest to the newest (e.g., in reverse order of prior events)."
	^(self droppedEventStreamOfSize: 0) contents! !

!ScriptEvent methodsFor: 'accessing' stamp: 'ar 1/10/2008 17:45'!
withDroppedEventsDo: aBlock
	"Evaluate aBlock with the receiver and all dropped events from the oldest to the newest."
	self priorEvent ifNotNilDo:[:evt| evt withDroppedEventsDo: aBlock].
	^aBlock value: self! !

ScriptEvent subclass: #ScriptNullEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptNullEvent methodsFor: 'testing' stamp: 'ar 1/8/2008 17:45'!
coalesceEvents
	"Answer true if events like the receiver should be coalesced or not"
	^true
! !

ScriptEvent subclass: #ScriptTickEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptTickEvent methodsFor: 'testing' stamp: 'ar 1/8/2008 17:47'!
coalesceEvents
	"Answer true if events like the receiver should be coalesced or not"
	^true
! !

Link weakSubclass: #ScriptHandlerLink
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptHandlerLink class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 19:09'!
new
	^self new: 1! !

!ScriptHandlerLink class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 19:10'!
on: anObject
	^(self new: 1) value: anObject! !

!ScriptHandlerLink methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:10'!
value
	^self at: 1! !

!ScriptHandlerLink methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:10'!
value: anObject
	self at: 1 put: anObject! !

!UndefinedObject methodsFor: '*Scripting-Extensions' stamp: 'ar 1/23/2005 18:47'!
on: eventName notify: anObserver
	"ignore"! !

!UndefinedObject methodsFor: '*Scripting-Extensions' stamp: 'ar 1/23/2005 18:48'!
on: eventName unNotify: anObserver
	"ignore"! !

!UndefinedObject methodsFor: '*Scripting-Extensions' stamp: 'ar 5/13/2005 11:22'!
waitTick
	"ignore"
	^nil! !

!Semaphore methodsFor: '*Scripting-override' stamp: 'ar 10/6/2007 11:07'!
critical: mutuallyExcludedBlock			
	"Evaluate mutuallyExcludedBlock only if the receiver is not currently in
	the process of running the critical: message. If the receiver is, evaluate
	mutuallyExcludedBlock after the other critical: message is finished."
	| blockValue process caught |
	process := Processor activeProcess.
	process isSystemProcess ifFalse:[process beginCritical: self].

	caught := false.
	"The following works only correctly in combination with Process>>terminate."
	[
		caught := true. 
		self primitiveWait.
		blockValue := mutuallyExcludedBlock value
	] ensure: [
		caught ifTrue: [self signal].
		process isSystemProcess ifFalse:[process endCritical: self].
	].
	^blockValue
! !

!Semaphore methodsFor: '*Scripting-Preload' stamp: 'ar 2/23/2005 16:19'!
primitiveWait
	"Primitive. The active Process must receive a signal through the receiver 
	before proceeding. If no signal has been sent, the active Process will be 
	suspended until one is sent. Essential. See Object documentation 
	whatIsAPrimitive."
	<primitive: 86>
	self primitiveFailed
	"excessSignals>0  
		ifTrue: [excessSignals := excessSignals-1]  
		ifFalse: [self addLastLink: Processor activeProcess suspend]"! !

!Semaphore methodsFor: '*Scripting-override' stamp: 'ar 3/6/2002 00:44'!
wait
	"The active Process must receive a signal through the receiver 
	before proceeding. If no signal has been sent, the active Process will be 
	suspended until one is sent. Essential. See Object documentation 
	whatIsAPrimitive."
	| process |
	(process := Processor activeProcess) isSystemProcess
		ifTrue:[^self primitiveWait].
	process beginWaitOn: self.
	self primitiveWait.
	process endWaitOn: self.! !

!Delay methodsFor: '*Scripting-Preload' stamp: 'ar 8/31/2007 17:25'!
primitiveWait
	"Schedule this Delay, then wait on its semaphore. The current process will be suspended for the amount of time specified when this Delay was created."
	self schedule.
	[delaySemaphore primitiveWait] ifCurtailed:[self unschedule].! !

!Delay methodsFor: '*Scripting-Extensions' stamp: 'ar 8/22/2001 14:18'!
setDelayDuration: millisecondCount
	delayDuration := millisecondCount asInteger max: 0.! !

!BlockContext methodsFor: '*Scripting-Extensions' stamp: 'ar 3/26/2007 22:44'!
asAsyncScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	| nArgs selector |
	nArgs := self numArgs.
	nArgs = 0 ifTrue:[selector := #evaluate:].
	nArgs = 1 ifTrue:[selector := #evaluate:with:].
	nArgs = 2 ifTrue:[selector := #evaluate:with:with:].
	nArgs = 3 ifTrue:[selector := #evaluate:with:with:with:].
	nArgs = 4 ifTrue:[selector := #evaluate:with:with:with:with:].
	selector ifNil:[^super asAsyncScriptMessageIn: scriptPerformer arguments: args].
	^Processor activeIsland asyncScriptMessageSend 
		receiver: scriptPerformer 
		selector: selector
		arguments: {self},args! !

!BlockContext methodsFor: '*Scripting-Extensions' stamp: 'ar 4/21/2005 22:13'!
asScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	| nArgs selector |
	nArgs := self numArgs.
	nArgs = 0 ifTrue:[selector := #evaluate:].
	nArgs = 1 ifTrue:[selector := #evaluate:with:].
	nArgs = 2 ifTrue:[selector := #evaluate:with:with:].
	nArgs = 3 ifTrue:[selector := #evaluate:with:with:with:].
	nArgs = 4 ifTrue:[selector := #evaluate:with:with:with:with:].
	selector ifNil:[^super asAsyncScriptMessageIn: scriptPerformer arguments: args].
	^ScriptMessageSend 
		receiver: scriptPerformer 
		selector: selector
		arguments: {self},args! !

!BlockContext methodsFor: '*Scripting-Extensions' stamp: 'ar 4/21/2005 22:11'!
isScriptDescriptorIn: aScript
	"Convert the receiver into an appropriate ScriptMessageSend"
	| nArgs selector |
	nArgs := self numArgs.
	nArgs = 0 ifTrue:[selector := #evaluate:].
	nArgs = 1 ifTrue:[selector := #evaluate:with:].
	nArgs = 2 ifTrue:[selector := #evaluate:with:with:].
	nArgs = 3 ifTrue:[selector := #evaluate:with:with:with:].
	nArgs = 4 ifTrue:[selector := #evaluate:with:with:with:with:].
	selector ifNil:[^false].
	^aScript selector == selector and:[(aScript arguments at: 1) == self].! !

Exception subclass: #DropAllButLastEvent
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!DropAllButLastEvent commentStamp: '<historical>' prior: 0!
This exception can be signaled by an event handler to drop all but the last event currently executed by a handler. In other words, a handler of the form:

onFoo: arg
	<on: foo>
	DropAllButLastEvent signal.
	Transcript cr; show: arg.

that is invoked using:

	self signal: #foo with: 1.
	self signal: #foo with: 2.
	self signal: #foo with: 3.

would first be executed with argument 1 but upon signaling DropAllButLastEvent re-evaluate the handler with *only* the last event.
!

!DropAllButLastEvent methodsFor: 'handling' stamp: 'ar 1/16/2008 14:07'!
defaultAction
	"If there is no handler for this exception, simply keep going"
	self resume.! !

!DropAllButLastEvent methodsFor: 'testing' stamp: 'ar 1/16/2008 14:07'!
isResumable
	"This exception is resumable"
	^true! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 12:08'!
asAsyncScriptMessageIn: scriptPerformer arguments: args
	^self error: 'Invalid script descriptor: ', self printString! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/24/2001 21:20'!
asScriptMessageIn: scriptPerformer arguments: args
	^self error: 'Invalid script descriptor: ', self printString! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/14/2003 04:36'!
asTickMessageIn: scriptPerformer arguments: args
	^self error: 'Invalid script descriptor: ', self printString! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:34'!
catchAnyOf: eventSpec during: aBlock
	"Catch any of the events in eventSpec during the evaluation of aBlock.
	DETAILS: All of the events triggered will be passed into the shared queue. If there are multiple events triggered, the last one will be passed upon return but any others (or older ones) are still accessible as dropped events."
	aBlock numArgs = 0 ifFalse:[^self error:'0 argument block expected'].
	eventSpec size = 0 ifTrue:[^aBlock value].
	^Processor activeProcess
		observe: eventSpec 
		semaphore: nil 
		buffer: ScriptEventBuffer new
		during:[:handler|
			aBlock value.
			handler nextEvent]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 8/22/2001 00:42'!
clock
	"Return the clock driving the receiver"
	^Processor activeProcess clock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 22:29'!
evaluate: aBlock
	^aBlock value! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 22:29'!
evaluate: aBlock with: arg1
	^aBlock value: arg1! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 22:31'!
evaluate: aBlock with: arg1 with: arg2
	^aBlock value: arg1 value: arg2! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/21/2005 22:11'!
evaluate: aBlock with: arg1 with: arg2 with: arg3
	^aBlock value: arg1 value: arg2 value: arg3! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/21/2005 22:12'!
evaluate: aBlock with: arg1 with: arg2 with: arg3 with: arg4
	^aBlock valueWithArguments: {arg1. arg2. arg3. arg4}! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/28/2002 16:33'!
findScript: scriptDescriptor
	| myScripts script |
	myScripts := self myScripts.
	1 to: myScripts size do:[:i|
		script := myScripts at: i.
		(scriptDescriptor isScriptDescriptorIn: script) ifTrue:[^script].
	].
	^nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/19/2002 01:07'!
findScript: scriptDescriptor withArguments: args
	| myScripts script |
	myScripts := self myScripts.
	1 to: myScripts size do:[:i|
		script := myScripts at: i.
		(scriptDescriptor isScriptDescriptorIn: script) ifTrue:[
			script arguments = args ifTrue:[^script].
		].
	].
	^nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:34'!
forAnyOf: eventSpec do: aBlock
	"Evaluate aBlock whenever any of the events in eventSpec is triggered."
	^self forAnyOf: eventSpec do: aBlock until:[false]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:34'!
forAnyOf: eventSpec do: aBlock until: stopCondition
	"Evaluate aBlock whenever any of the events in eventSpec is triggered.
	DETAILS: All of the events triggered will be passed into the shared queue. If there are multiple events triggered, the last one will be passed into the block but any others (or older ones) are still accessible as dropped events."
	| nArgs |
	(nArgs := aBlock numArgs) > 2 ifTrue:[^self error:'0-2 argument block expected'].
	eventSpec size = 0 ifTrue:[^nil].
	^Processor activeProcess
		observe: eventSpec
		semaphore: Semaphore new 
		buffer: ScriptEventBuffer new
		during:[:handler|
			[Processor activeProcess waitProcessOn: handler semaphore.
			self privateSendEvent: handler nextEvent to: aBlock args: nArgs.
			stopCondition value] whileFalse.
			nil].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
forAnyOf: eventSpec until: stopCondition do: aBlock			 
	"Evaluate aBlock whenever any of the events in eventSpec is triggered"
	^self forAnyOf: eventSpec do: aBlock until: stopCondition! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/20/2001 23:48'!
forEach: eventName do: aBlock
	"Evaluate aBlock whenever the event is triggered."
	^self forEach: eventName do: aBlock until: [false] queueSize: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/20/2001 23:48'!
forEach: eventName do: aBlock until: stopName
	^self waitUntil: stopName forEach: eventName do: aBlock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 19:35'!
forEach: eventName do: aBlock until: stopCondition queueSize: queueSize
	"Evaluate the given block whenever eventName is triggered"
	| nArgs |
	(nArgs := aBlock numArgs) > 2 ifTrue:[^self error:'0-2 argument block expected'].
	^Processor activeProcess
		observe: {self. eventName}
		semaphore: nil 
		buffer: nil
		during:[:handler|
			[Processor activeProcess waitProcessOn: handler semaphore.
			self privateSendEvent: handler nextEvent to: aBlock args: nArgs.
			stopCondition value] whileFalse.
			nil]! !

!Object methodsFor: '*Scripting-Extensions' stamp: ''!
forEach: eventName do: aBlock untilAnyOf: stopEvents
	^self waitUntilAnyOf: stopEvents forEach: eventName do: aBlock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 2/6/2002 16:17'!
forEach: eventName until: stopCondition do: aBlock
	"Evaluate the given block whenever eventName is triggered"
	^self forEach: eventName do: aBlock until: stopCondition queueSize: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
forEach: loopEvents untilAnyOf: stopEvents do: aBlock
	^self waitUntilAnyOf: stopEvents forEach: loopEvents do: aBlock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 20:21'!
handlesAnyOf: eventList
	| map |
	map := self myEventMap ifNil:[^false].
	eventList do:[:evtName| (self handlesEvent: evtName map: map) ifTrue:[^true]].
	^false! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 20:22'!
handlesEvent: eventName
	^self handlesEvent: eventName map: (self myEventMap ifNil:[^false]).! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 14:43'!
handlesEvent: eventName map: map
	| list |
	list := map at: eventName ifAbsent:[^false].
	1 to: list size do:[:i|
		(list at: i) ifNotNil:[^true].
	].
	self removeEvent: eventName map: map.
	^false! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2002 20:46'!
isEventTracer
	^false! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 2/17/2002 19:11'!
isScript
	^false! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 2/6/2004 17:42'!
isSpecialSystemScript: scriptDescriptor
	"A fall-back for special system scripts which need to be handled in slightly different ways"
	^scriptDescriptor == #runTicking:at:! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/12/2005 20:55'!
makeTicking: scriptDescriptor withArguments: args at: frequency
	"If no trigger is given, the script is evaluated exactly once"
	| script |
	script := self findScript: scriptDescriptor withArguments: args.
	(script == nil or:[script isTickingScript not]) ifTrue:[
		"Create a new script"
		script := scriptDescriptor asTickMessageIn: self arguments: args.
		script eventSpec: (Array with: self with: #ticking).
	] ifFalse:[
		script pauseScript.
	].
	script frequency: frequency.
	self privateAddScript: script.
	self signal: #scriptStatusChanged with: script selector.
	^script! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 21:53'!
myEventMap
	^self myEventRegistry at: self ifAbsent:[nil]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 21:56'!
myEventMap: anEventMap
	anEventMap 
		ifNil:[self myEventRegistry removeKey: self ifAbsent:[]]
		ifNotNil:[self myEventRegistry at: self put: anEventMap].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 21:55'!
myEventRegistry
	^Smalltalk at: #TestingEventMap  ifAbsentPut:[WeakIdentityKeyDictionary new].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 21:54'!
myScriptRegistry
	^Smalltalk at: #TestingScriptRegistry  ifAbsentPut:[WeakIdentityKeyDictionary new].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 17:15'!
myScripts
	^self myScriptRegistry at: self ifAbsent:[#()]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 21:56'!
myScripts: scripts
	scripts
		ifNil:[self myScriptRegistry removeKey: self ifAbsent:[]]
		ifNotNil:[self myScriptRegistry  at: self put: scripts].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/20/2001 20:36'!
newEventMap
	^IdentityDictionary new! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/9/2004 17:19'!
on: eventName notify: anObserver
	"Notify the observer when the given event is triggered."
	| map list |
	eventName ifNil:[^self].
	map := self myEventMap.
	map ifNil:[self myEventMap: (map := self newEventMap)].
	list := map at: eventName ifAbsentPut:[WeakArray new: 1].
	1 to: list size do:[:i|
		(list at: i) ifNil:[^list at: i put: anObserver].
	].
	map at: eventName put: (list copyWith: anObserver).
	^anObserver
! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/9/2004 17:20'!
on: eventName traceWith: aTraceObserver
	"Start tracing events when the given event name is triggered.
	The provided trace observer will replace any existing trace observer for this event.
	If nil, the trace observer is removed."
	| map list |
	eventName ifNil:[^self].
	map := self myEventMap.
	map ifNil:[self myEventMap: (map := self newEventMap)].
	list := map at: eventName ifAbsentPut:[WeakArray new: 1].
	"tracer is always the first entry"
	(list at: 1) ifNil:[^list at: 1 put: aTraceObserver].
	(list at: 1) isEventTracer ifTrue:[^list at: 1 put: aTraceObserver].
	2 to: list size do:[:i|
		(list at: i) ifNil:[
			list at: i put: (list at: 1).
			^list at: 1 put: aTraceObserver].
	].
	map at: eventName put: (WeakArray with: aTraceObserver), list.
	^aTraceObserver
! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/9/2004 17:20'!
on: eventName unNotify: anObserver
	"Notify the observer when the given event is triggered."
	| map list |
	map := self myEventMap ifNil:[^self].
	list := map at: eventName ifAbsent:[^self].
	list := list reject:[:any| any == nil or:[any == anObserver]].
	list size = 0 
		ifTrue:[self removeEvent: eventName map: map]
		ifFalse:[map at: eventName put: list].
! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/19/2002 23:47'!
pauseScript: scriptDescriptor
	| myScripts script |
	myScripts := self myScripts.
	1 to: myScripts size do:[:i|
		script := myScripts at: i.
		(scriptDescriptor isScriptDescriptorIn: script) ifTrue:[
			script pauseScript.
		].
	].
	^script! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 19:35'!
primitiveWaitUntil: eventName
	"Wait until an event with the given name is signaled."
	^Processor activeProcess
		observe: {self. eventName}
		semaphore: nil 
		buffer: nil
		during:[:handler|
			handler semaphore primitiveWait.
			handler nextEvent
		].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 19:36'!
primitiveWaitUntilAnyOf: eventSpec
	"Wait until any of the events in eventSpec are triggered."
	eventSpec size = 0 ifTrue:[^nil].
	^Processor activeProcess
		observe: eventSpec
		semaphore: Semaphore new 
		buffer: ScriptEventBuffer new
		during:[:handler|
			handler semaphore primitiveWait.
			handler nextEvent
		].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 19:19'!
privateAddScript: aScript
	| scriptList |
	aScript receiver == self ifFalse:[^self error:'Not my script'].
	scriptList := self myScripts.
	scriptList size = 0
		ifTrue:[scriptList := Array with: aScript] 
		ifFalse:[scriptList := scriptList copyWith: aScript].
	self myScripts: scriptList! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 2/2/2006 17:54'!
privateRemoveScript: aScript
	| scriptList |
	scriptList := self myScripts.
	scriptList := scriptList reject:[:s| s == aScript].
	scriptList isEmpty ifTrue:[scriptList := nil].
	self myScripts: scriptList! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/20/2001 20:37'!
privateSendEvent: anEvent to: aBlock args: nArgs
	"Private. Send an event to the block given the number of block arguments"
	nArgs = 0 ifTrue:[^aBlock value].
	nArgs = 1 ifTrue:[^aBlock value: anEvent].
	nArgs = 2 ifTrue:[^aBlock value: anEvent value: self].
	self error:'0-2 argument block expected'! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 11/13/2004 15:31'!
privateSignalEvent: anEvent
	"Signal the occurance of anEvent"
	| map list msg |
	map := self myEventMap ifNil:[^self].
	list := map at: anEvent selector ifAbsent:[^self].
	1 to: list size do:[:i|
		(msg := list at: i) ifNotNil:[msg valueWithEvent: anEvent clone]].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 11/13/2004 15:31'!
privateSignalSynchronousEvent: anEvent
	"Signal the occurance of anEvent"
	| map list msg |
	map := self myEventMap ifNil:[^self].
	list := map at: anEvent selector ifAbsent:[^self].
	1 to: list size do:[:i|
		(msg := list at: i) ifNotNil:[msg synchronousValueWithEvent: anEvent clone]].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'jcg 2/23/2007 02:00'!
removeAllEvents
	self myEventMap: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 1/2/2007 22:41'!
removeEvent: eventName
	self removeEvent: eventName map: (self myEventMap ifNil:[^self]).! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 20:33'!
removeEvent: eventName map: eventMap
	eventMap removeKey: eventName ifAbsent:[].
	eventMap size = 0 ifTrue:[self myEventMap: nil].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 2/24/2004 01:34'!
resumeScript: scriptDescriptor
	| myScripts script |
	myScripts := self myScripts.
	1 to: myScripts size do:[:i|
		script := myScripts at: i.
		(scriptDescriptor isScriptDescriptorIn: script) ifTrue:[
			script resumeScript.
			^script
		].
	].
	^nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 12:11'!
runScript: scriptDescriptor
	^self runScript: scriptDescriptor withArguments: #() when: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 14:19'!
runScript: aScriptDescriptor when: aTrigger
"Evaluate the given script whenever the trigger becomes active. The script will be run in the context of the current island. The script will run in the calling process, and will finish before the calling #signal: method returns. Unlike #startScript:, #runScript: does not need a ScriptScheduler running in order to work correctly"

	^self runScript: aScriptDescriptor withArguments: #() when: aTrigger! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 12:11'!
runScript: scriptDescriptor withArguments: args
	^self runScript: scriptDescriptor withArguments: args when: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 14:18'!
runScript: aScriptDescriptor withArguments: args when: aTrigger
"Evaluate the given script whenever the trigger becomes active. The script will be run in the context of the current island. The script will run in the calling process, and will finish before the calling #signal: method returns. Unlike #startScript:, #runScript: does not need a ScriptScheduler running in order to work correctly"

	| script spec |
	script := aScriptDescriptor asScriptMessageIn: self arguments: args.
	self privateAddScript: script.
	aTrigger ifNil:[^script valueWithEvent: nil].
	spec := aTrigger asEventSpecificationIn: self.
	spec ifNil:[^script valueWithEvent: nil].
	1 to: spec size by: 2 do:[:i|
		(spec at: i) on: (spec at: i+1) notify: script.
	].
	script eventSpec: spec.
	^script! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 11/5/2007 19:15'!
runTicking: aMessage at: frequency
	"Run the given script ticking at the provided frequency"
	| delay script |
	script := Processor activeProcess.
	frequency < 1000 
		ifTrue:[delay := Delay forMilliseconds: (1000 / frequency) asInteger].
	[true] whileTrue:[
		"fire script"
		aMessage valueWithEvent: script scheduler tickEvent.
		"wait on delay"
		delay ifNotNil:[delay wait].
		self waitTick.
	].
! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/20/2003 12:29'!
send: aSelector to: anObject
	^anObject perform: aSelector! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/20/2003 12:29'!
send: aSelector to: anObject with: arg1
	^anObject perform: aSelector with: arg1! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/20/2003 12:29'!
send: aSelector to: anObject with: arg1 with: arg2
	^anObject perform: aSelector with: arg1 with: arg2! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/20/2003 12:29'!
send: aSelector to: anObject withArguments: args
	^anObject perform: aSelector withArguments: args! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:24'!
signal: eventName
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName).! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:25'!
signal: eventName with: arg1
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName arguments: (Array with: arg1))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:25'!
signal: eventName with: arg1 with: arg2
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName arguments: (Array with: arg1 with: arg2))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:26'!
signal: eventName with: arg1 with: arg2 with: arg3
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName arguments: (Array with: arg1 with: arg2 with: arg3))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:26'!
signal: eventName with: arg1 with: arg2 with: arg3 with: arg4
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName arguments: (Array with: arg1 with: arg2 with: arg3 with: arg4))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:26'!
signal: eventName withArguments: anArray
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptEvent new setSender: self selector: eventName arguments: anArray)! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:28'!
signalChanged: eventName from: oldValue to: newValue
"Signal the occurence of eventName, which specifies a change from oldValue to newValue. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self signalEvent: (ScriptChangeEvent new setSender: self selector: eventName arguments: (Array with: newValue with: oldValue))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:28'!
signalEvent: anEvent
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers created with #runScript: will run in the current process, and are guaranteed to have completed once this method returns. The handlers created with #startScript: will be run in a new process controlled by the ScriptScheduler that created them, and wll run eventually"
	^self privateSignalEvent: anEvent! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:30'!
signalSynchronousEvent: anEvent
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self privateSignalSynchronousEvent: anEvent! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 1/8/2008 17:46'!
signalTick: eventName
	"Signal the occurance of anEvent"
	^self signalEvent: (ScriptTickEvent new setSender: self selector: eventName arguments: #())
! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/24/2001 21:26'!
startScript: scriptDescriptor
	"If no trigger is given, the script is evaluated exactly once"
	^self startScript: scriptDescriptor withArguments: #() when: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 14:09'!
startScript: aScriptDescriptor when: aTrigger
"Evaluate the given script whenever the trigger becomes active. The script will be run in the context of the current island. If it is triggered by #signal: (the normal case) it will be started in a new process under the controll of the current ScriptScheduler. If triggered by #synchronousSignal:, it will run in the calling process, like a script created by #runScript:.. If running in the absence of a ScriptScheduler, use #runScript: instead"

	^self startScript: aScriptDescriptor withArguments: #() when: aTrigger! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/24/2001 21:25'!
startScript: scriptDescriptor withArguments: args
	"If no trigger is given, the script is evaluated exactly once"
	^self startScript: scriptDescriptor withArguments: args when: nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 14:09'!
startScript: aScriptDescriptor withArguments: args when: aTrigger
"Evaluate the given script whenever the trigger becomes active. The script will be run in the context of the current island. If it is triggered by #signal: (the normal case) it will be started in a new process under the controll of the current ScriptScheduler. If triggered by #synchronousSignal:, it will run in the calling process, like a script created by #runScript:.. If running in the absence of a ScriptScheduler, use #runScript: instead"

	| script spec |
	script := aScriptDescriptor asAsyncScriptMessageIn: self arguments: args.
	self privateAddScript: script.
	self signal: #scriptStatusChanged with: script selector.
	aTrigger ifNil:[^script valueWithEvent: ScriptNullEvent new].
	spec := aTrigger asEventSpecificationIn: self.
	spec ifNil:[^script valueWithEvent: ScriptNullEvent new].
	"Note: Normally, we should be able to just do a pauseScript/resumeScript here. 
	However, this would dump any existing event queue in the async scripts which
	may already contain events."
	1 to: spec size by: 2 do:[:i|
		(spec at: i) on: (spec at: i+1) notify: script.
	].
	script eventSpec size = 0
		ifTrue:[script eventSpec: spec]
		ifFalse:[script eventSpec: script eventSpec, spec].
	^script! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/14/2003 04:45'!
startTicking: scriptDescriptor at: frequency
	"If no trigger is given, the script is evaluated exactly once"
	^self startTicking: scriptDescriptor withArguments: #() at: frequency! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2005 11:35'!
startTicking: scriptDescriptor withArguments: args at: frequency
	"If no trigger is given, the script is evaluated exactly once"
	| script |
	script := self makeTicking: scriptDescriptor withArguments: args at: frequency.
	script resumeScript.
	self signal: #scriptStatusChanged with: script selector.
	^script! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 9/17/2008 15:49'!
stopAllScriptsTriggeredBy: receiver
	"Stop all the scripts that are triggered by the given receiver. I.e.,
		self startScript: #onFoo when: {rcvr. #foo}.
		self startScript: #onBar when: {rcvr. #bar}.
	...
		self stopAllScriptsTriggeredBy: rcvr.
	will stop the scripts registered earlier."
	| scripts |
	scripts := (self myScripts ifNil:[^self]) select:[:each| each isTriggeredBy: receiver].
	scripts do:[:sc| self stopScript: sc].
	^scripts! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 10/23/2006 18:57'!
stopScript: scriptDescriptor
	| myScripts script |
	myScripts := self myScripts.
	1 to: myScripts size do:[:i|
		script := myScripts at: i.
		(scriptDescriptor isScriptDescriptorIn: script) ifTrue:[
			script terminateScript.
		].
	].
	"Don't return 'script' since it may not have been the script we stopped"
	^nil! !

!Object methodsFor: '*Scripting-Extensions' stamp: ''!
stopScript: aScript when: anEvent
	self startScript: #stopScript: withArguments:{aScript} when: anEvent! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/29/2002 22:54'!
stopScripts: scriptList
	scriptList do:[:script| self stopScript: script].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 00:11'!
syncScript
	"Sync the current script, e.g., suspend it until all pending scripts have completed"
	Processor activeProcess scheduler syncList primitiveWait.! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:31'!
synchronousSignal: eventName
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self signalSynchronousEvent: (ScriptEvent new setSender: self selector: eventName).! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:31'!
synchronousSignal: eventName with: arg1
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self signalSynchronousEvent: 
		(ScriptEvent new 
			setSender: self 
			selector: eventName
			arguments: (Array with: arg1))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:31'!
synchronousSignal: eventName with: arg1 with: arg2
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self signalSynchronousEvent: 
		(ScriptEvent new 
			setSender: self 
			selector: eventName
			arguments: (Array with: arg1 with: arg2))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:31'!
synchronousSignal: eventName with: arg1 with: arg2 with: arg3
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self signalSynchronousEvent: 
		(ScriptEvent new 
			setSender: self 
			selector: eventName
			arguments: (Array with: arg1 with: arg2 with: arg3))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'mtf 6/22/2011 13:31'!
synchronousSignal: eventName with: arg1 with: arg2 with: arg3 with: arg4
"Signal the occurance of anEvent. Run any handlers for the event in the context of the island they were created in. The handlers will run in the current process, and are guaranteed to have completed once this method returns, regardless of whether they were created with #startScript: or #runScript:"
	^self signalSynchronousEvent: 
		(ScriptEvent new 
			setSender: self 
			selector: eventName
			arguments: (Array with: arg1 with: arg2 with: arg3 with: arg4))! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/19/2002 23:30'!
wait: timeOut
	"Wait for the given number of seconds"
	timeOut > 0 ifTrue:[(Delay forMilliseconds: (timeOut * 1000) asInteger) wait].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 7/21/2001 21:38'!
waitOn: aWaitable
	^Processor activeProcess waitProcessOn: aWaitable.! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 5/26/2005 18:23'!
waitTick
	^Processor activeProject waitUntil: #tick.! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/14/2004 15:14'!
waitUntil: eventName
	"Wait until an event with the given name is signaled.
	[ar 3/14/2004] Something is wrong with the semantics here. While it is perfectly fine for a player to wait until something happens inside himself, e.g., 'self waitUntil: #mouseDown', it seems Very Wrong Indeed to ask another player to 'container waitUntil: #mouseDown' since it isn't the container who is waiting but rather the sender of the request. 
	So what might be a good representation for this? We know it's always the sender who is waiting and never the guy initiating the event... maybe self waitUntil: #mouseDown in: container? Or even 'self waitUntil: container's mouseDown'? Grrrrr... so how do we distinguish the name of some event from the name of a method???"
	| script |
	^(script := Processor activeProcess)
		observe: {self. eventName}
		semaphore: nil 
		buffer: nil
		during:[:handler|
			script waitProcessOn: handler semaphore.
			handler nextEvent
		]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
waitUntil: eventName do: aBlock
	"Repeatedly evaluate aBlock until the given event is signaled."
	^self waitUntil: eventName forAnyOf: #() do: aBlock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
waitUntil: stopEvent forAnyOf: eventSpec do: aBlock
	"Evaluate aBlock for each of the events generated from event spec. Stop if stopEvent occurs in stopObject. See comment at the end for details."
	| sema nArgs script loopHandler |
	(nArgs := aBlock numArgs) > 2 ifTrue:[^self error:'0-2 argument block expected'].
	^(script := Processor activeProcess)
		observe: {self. stopEvent}
		semaphore: nil
		buffer: nil
		during:[:stopHandler|
			sema := stopHandler semaphore.
			loopHandler := script
				observe: eventSpec
				semaphore: sema 
				buffer: ScriptEventBuffer new.
			 [sema ifNotNil:[script waitProcessOn: sema]. "until something happens"
			stopHandler size = 0] whileTrue:
				[self privateSendEvent: loopHandler nextEvent to: aBlock args: nArgs].
			stopHandler nextEvent].
	"DETAILS: This method establishes two separate scopes to handle stop and loop events independently. An alternative way would be to use something like:
		self forAnyOf: {
			self. stopEvent.
			loopObject. loopEvent.
		} do:[:evt|
			evt selector == stopEvent
				ifTrue:[doStop := true]
				ifFalse:[aBlock value: evt].
		] until:[doStop]
	but this has some complicated implications when the block takes a long time to complete. In this case, the stop event could get lost (if too many loop events come in afterwards) and it would also increase the complexity for the case if any loop events come in after the stop event. The way it is done above avoids these problems completely."! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
waitUntil: stopEvent forEach: loopEvent do: aBlock
	^self waitUntil: stopEvent forEach: loopEvent in: self do: aBlock! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
waitUntil: stopEvent forEach: loopEvent in: loopObject do: aBlock
	"Evaluate aBlock for each of the loop events generated by loopObject. Stop if stopEvent occurs in stopObject. See comment at the end for details."
	| sema nArgs loopHandler script |
	(nArgs := aBlock numArgs) > 2 ifTrue:[^self error:'0-2 argument block expected'].
	^(script := Processor activeProcess)
		observe: {self. stopEvent}
		semaphore: nil
		buffer: nil
		during:[:stopHandler|
			sema := stopHandler semaphore.
			loopHandler := script observe: {loopObject. loopEvent} semaphore: sema.
			[script waitProcessOn: sema. "until something happens"
			stopHandler size = 0] whileTrue:
				[self privateSendEvent: loopHandler nextEvent to: aBlock args: nArgs].
			stopHandler nextEvent].
	"DETAILS: This method establishes two separate scopes to handle stop and loop events independently. An alternative way would be to use something like:
		self forAnyOf: {
			self. stopEvent.
			loopObject. loopEvent.
		} do:[:evt|
			evt selector == stopEvent
				ifTrue:[doStop := true]
				ifFalse:[aBlock value: evt].
		] until:[doStop]
	but this has some complicated implications when the block takes a long time to complete. In this case, the stop event could get lost (if too many loop events come in afterwards) and it would also increase the complexity for the case if any loop events come in after the stop event. The way it is done below avoids these problems completely."! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'bf 7/25/2005 15:09'!
waitUntil: event timeout: milliseconds
	"Wait until event was triggered, but return after the specified amount of milliseconds even if no event happened"
	^self waitUntilAnyOf: {self. event} timeout: milliseconds! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 19:48'!
waitUntilAllOf: eventSpec
	"Wait until all of the events in event spec are triggered.
	Return an array of events corresponding the event specification."
	| tempList anObject selector handler script |
	eventSpec size = 0 ifTrue:[^#()].
	"Since we need to have access to all the semaphores,
	create and remember each handler individually."
	script := Processor activeProcess.
	tempList := Array new: eventSpec size // 2.
	1 to: eventSpec size // 2 do:[:i|
		anObject := eventSpec at: 2*i-1.
		selector := eventSpec at: 2*i.
		tempList at: i put: (script observe: {anObject. selector}).
	].
	handler := tempList at: 1. "the outermost one"
	[
		tempList do:[:h| script waitProcessOn: h semaphore].
		1 to: tempList size do:[:i|
			tempList at: i put: (tempList at: i) nextEvent].
	] ensure:[
		script releaseObserver: handler.
	].
	^tempList! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 4/23/2002 19:49'!
waitUntilAnyOf: eventSpec
	"Wait until any of the events in eventSpec are triggered."
	| script |
	eventSpec size = 0 ifTrue:[^nil].
	^(script := Processor activeProcess)
		observe: eventSpec
		semaphore: Semaphore new 
		buffer: ScriptEventBuffer new
		during:[:handler|
			script waitProcessOn: handler semaphore.
			handler nextEvent]! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 3/29/2005 21:35'!
waitUntilAnyOf: stopEvents forEach: loopEvents do: aBlock
	"Evaluate aBlock for each of the loop events generated by loopObject. Stop if stopEvent occurs in stopObject. See comment at the end for details."
	| sema nArgs loopHandler script |
	(nArgs := aBlock numArgs) > 2 ifTrue:[^self error:'0-2 argument block expected'].
	^(script := Processor activeProcess)
		observe: stopEvents
		semaphore: Semaphore new
		buffer: ScriptEventBuffer new
		during:[:stopHandler|
			sema := stopHandler semaphore.
			loopHandler := script observe: loopEvents semaphore: sema.
			[script waitProcessOn: sema. "until something happens"
			stopHandler buffer size = 0] whileTrue:
				[self privateSendEvent: loopHandler nextEvent to: aBlock args: nArgs].
			stopHandler nextEvent].! !

!Object methodsFor: '*Scripting-Extensions' stamp: 'ar 8/30/2007 16:20'!
waitUntilAnyOf: eventSpec timeout: milliseconds
	"Wait until any of the events in eventSpec are triggered, but return after the specified amount of milliseconds even if no event happened"
	| script delay |
	eventSpec size = 0 ifTrue:[^nil].
	^(script := Processor activeProcess)
		observe: eventSpec
		semaphore: Semaphore new 
		buffer: ScriptEventBuffer new
		during:[:handler|
			delay := Delay timeoutSemaphore: handler semaphore
				afterMSecs: (milliseconds max: 0).
			[script waitProcessOn: handler semaphore] ensure:[delay unschedule].
			handler nextEvent]! !

Object subclass: #ScriptEventBuffer
	instanceVariableNames: 'firstEvent lastEvent size'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptEventBuffer class methodsFor: 'instance creation' stamp: 'ar 1/30/2005 11:57'!
new
	^self basicNew initialize! !

!ScriptEventBuffer methodsFor: 'initialize' stamp: 'ar 7/23/2001 22:44'!
flush
	firstEvent := lastEvent := nil.
	size := 0.! !

!ScriptEventBuffer methodsFor: 'initialize' stamp: 'ar 3/29/2005 21:34'!
initialize
	firstEvent := lastEvent := nil.
	size := 0.! !

!ScriptEventBuffer methodsFor: 'streaming' stamp: 'ar 4/25/2002 11:56'!
next
	| prev now next |
	size > 1 ifTrue:["READ THE CLASS COMMENT!!!!!!"
		prev := firstEvent.
		now := firstEvent priorEvent.
		firstEvent priorEvent: nil.
		[now == nil] whileFalse:[
			"Change link now->next to now->prev"
			next := now priorEvent.
			now priorEvent: prev.
			prev := now.
			now := next].
	].
	next := lastEvent.
	lastEvent := firstEvent := nil.
	size := 0.
	^next! !

!ScriptEventBuffer methodsFor: 'streaming' stamp: 'ar 3/31/2005 11:18'!
nextPut: anEvent
	size = 0 ifTrue:[
		firstEvent := lastEvent := anEvent.
	] ifFalse:[
		lastEvent priorEvent: anEvent. "READ THE CLASS COMMENT!!!!!!"
		lastEvent := anEvent.
	].
	size := size + 1.
! !

!ScriptEventBuffer methodsFor: 'accessing' stamp: 'ar 7/20/2001 21:13'!
size
	"Return the number of elements in the receiver"
	^size! !

Object subclass: #ScriptEventObserver
	instanceVariableNames: 'flags semaphore buffer nextHandler suspended signaler selector myFarRef'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:22'!
buffer
	^buffer ifNil:[buffer := ScriptEventBuffer new]! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:22'!
buffer: aBuffer
	buffer := aBuffer! !

!ScriptEventObserver methodsFor: 'testing' stamp: 'ar 4/24/2002 15:23'!
isSuspended
	^(flags bitAnd: 1) == 1! !

!ScriptEventObserver methodsFor: 'islands' stamp: 'ar 5/24/2005 15:49'!
myFarRef
	^myFarRef! !

!ScriptEventObserver methodsFor: 'islands' stamp: 'ar 5/24/2005 15:49'!
myFarRef: farRef
	myFarRef := farRef! !

!ScriptEventObserver methodsFor: 'streaming' stamp: 'ar 4/23/2002 20:35'!
next
	semaphore initSignals.
	^buffer ifNotNil:[buffer next]! !

!ScriptEventObserver methodsFor: 'streaming' stamp: 'ar 4/23/2002 20:52'!
nextEvent
	semaphore initSignals.
	^buffer ifNotNil:[buffer next]! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:27'!
nextHandler
	^nextHandler! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:28'!
nextHandler: aHandler
	nextHandler := aHandler! !

!ScriptEventObserver methodsFor: 'streaming' stamp: 'ar 4/23/2002 20:35'!
nextPut: anEvent
	self buffer nextPut: anEvent.
	semaphore signal.! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 20:36'!
noticeEvent: anEvent
	^self nextPut: anEvent! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'bf 11/11/2005 16:26'!
release
	suspended ifNotNil:[suspended resume].
	signaler on: selector unNotify: self.
	suspended := signaler := nil.! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'ar 4/23/2002 19:29'!
releaseTo: aHandler
	aHandler == self ifTrue:[^self].
	self release.
	^nextHandler ifNotNil:[nextHandler releaseTo: aHandler].! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'ar 4/24/2002 15:23'!
resume
	flags := flags bitClear: 1.! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'ar 4/24/2002 15:22'!
resumeSuspending: aHandler
	flags := 0.
	aHandler ifNotNil:[aHandler suspend].
	suspended := aHandler.! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:20'!
selector
	^selector! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:20'!
selector: anObject
	selector := anObject! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:21'!
semaphore
	^semaphore! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:21'!
semaphore: sema
	semaphore := sema! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:20'!
signaler
	^signaler! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:21'!
signaler: anObject
	signaler := anObject! !

!ScriptEventObserver methodsFor: 'streaming' stamp: 'ar 4/23/2002 20:36'!
size
	^buffer ifNil:[0] ifNotNil:[buffer size]! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'ar 4/24/2002 15:23'!
suspend
	flags := flags bitOr: 1.! !

!ScriptEventObserver methodsFor: 'initialize-release' stamp: 'ar 4/24/2002 15:19'!
suspend: aHandler
	aHandler ifNotNil:[aHandler suspend].
	suspended := aHandler.! !

!ScriptEventObserver methodsFor: 'accessing' stamp: 'ar 4/23/2002 19:28'!
suspendedHandler
	^suspended! !

!ScriptEventObserver methodsFor: 'streaming' stamp: 'ar 4/24/2002 15:23'!
valueWithEvent: event
	self isSuspended ifTrue:[^self].
	self buffer nextPut: event.
	semaphore signal.! !

Object subclass: #ScriptEventTraceEntry
	instanceVariableNames: 'event message script parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptEventTraceEntry methodsFor: 'accessing' stamp: 'ar 5/17/2002 20:21'!
event
	^event! !

!ScriptEventTraceEntry methodsFor: 'initialize' stamp: 'ar 5/17/2002 20:39'!
event: anEvent message: aMessage script: aScript parent: parentScript
	event := anEvent.
	message := aMessage.
	script := aScript.
	parent := parentScript.! !

!ScriptEventTraceEntry methodsFor: 'accessing' stamp: 'ar 5/17/2002 20:22'!
message
	^message! !

!ScriptEventTraceEntry methodsFor: 'accessing' stamp: 'ar 5/17/2002 20:39'!
parent
	^parent! !

!ScriptEventTraceEntry methodsFor: 'accessing' stamp: 'ar 5/17/2002 20:22'!
script
	^script! !

Object subclass: #ScriptEventTracer
	instanceVariableNames: 'traceLog traceEvent traceTarget traceDonePosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptEventTracer class methodsFor: 'instance creation' stamp: 'ar 5/17/2002 20:40'!
trace: eventName in: sender
	^self new trace: eventName in: sender! !

!ScriptEventTracer methodsFor: 'testing' stamp: 'ar 5/17/2002 21:01'!
isDone
	"Return true if no more events can be traced"
	| entry script |
	traceLog position = 0 ifTrue:[^false].
	traceDonePosition = traceLog size ifTrue:[^true].
	"See if any of the scripts found after traceDonePosition are still running"
	traceDonePosition = 0 ifTrue:[traceDonePosition := 1].
	traceLog position: traceDonePosition.
	[(entry := traceLog next) == nil] whileFalse:[
		script := entry script.
		(script notNil and:[script isDone]) ifFalse:[
			"Not yet finished"
			traceDonePosition := traceLog position.
			traceLog position: traceLog size.
		].
	].
	traceDonePosition := traceLog position.
	^true! !

!ScriptEventTracer methodsFor: 'testing' stamp: 'ar 5/17/2002 20:46'!
isEventTracer
	^true! !

!ScriptEventTracer methodsFor: 'tracing' stamp: 'ar 5/17/2002 20:45'!
stopTracing
	"For now, just uninstall - we might allow repeated evaluation later"
	traceTarget on: traceEvent traceWith: nil. "uninstall"! !

!ScriptEventTracer methodsFor: 'initialize' stamp: 'ar 5/17/2002 20:50'!
trace: eventName in: anObject
	traceLog := (ReadWriteStream on: (Array new: 100)).
	traceDonePosition := 0.
	traceEvent := eventName.
	traceTarget := anObject.
	traceTarget on: eventName traceWith: self.! !

!ScriptEventTracer methodsFor: 'tracing' stamp: 'ar 5/17/2002 20:38'!
traceEvent: anEvent message: aMessage script: aScript from: parentScript
	traceLog nextPut: (ScriptEventTraceEntry new
		event: anEvent message: aMessage script: aScript parent: parentScript).! !

!ScriptEventTracer methodsFor: 'tracing' stamp: 'ar 5/18/2002 01:08'!
valueWithEvent: anEvent
	"This initiates a new trace action"
	| script |
	script := Processor activeProcess.
	script eventTracer: self.
	script traceEvents: true.
	self traceEvent: anEvent message: (MessageSend receiver: traceTarget selector: traceEvent) script: Processor activeProcess from: nil.! !

Object subclass: #ScriptProcessLink
	instanceVariableNames: 'prevLink nextLink'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptProcessLink class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 21:13'!
on: aProcess
	^self new process: aProcess! !

!ScriptProcessLink class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 21:14'!
strongOn: aProcess
	^ScriptProcessLinkStrong new process: aProcess! !

!ScriptProcessLink class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 21:15'!
weakOn: aProcess
	^(ScriptProcessLinkWeak new: 1) process: aProcess! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:38'!
nextLink
	^nextLink! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:38'!
nextLink: aLink
	nextLink := aLink! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:38'!
prevLink
	^prevLink! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 7/14/2001 16:38'!
prevLink: aLink
	prevLink := aLink! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:14'!
process
	^self subclassResponsibility! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:14'!
process: aProcess
	^self subclassResponsibility! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 2/16/2002 20:53'!
value
	^self process! !

!ScriptProcessLink methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:11'!
value: aProcess
	^self process: aProcess! !

ScriptProcessLink subclass: #ScriptProcessLinkStrong
	instanceVariableNames: 'process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptProcessLinkStrong methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:13'!
process
	^process! !

!ScriptProcessLinkStrong methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:13'!
process: aProcess
	process := aProcess! !

ScriptProcessLink weakSubclass: #ScriptProcessLinkWeak
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptProcessLinkWeak class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 21:13'!
new
	^self new: 1! !

!ScriptProcessLinkWeak class methodsFor: 'instance creation' stamp: 'ar 4/23/2002 21:13'!
on: aProcess
	^(self new: 1) at: 1 put: aProcess; yourself! !

!ScriptProcessLinkWeak methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:15'!
process
	^self at: 1! !

!ScriptProcessLinkWeak methodsFor: 'accessing' stamp: 'ar 4/23/2002 21:15'!
process: aProcess
	self at: 1 put: aProcess! !

Object subclass: #ScriptScheduler
	instanceVariableNames: 'activeList scheduledList tickList activeScript interruptProtect myEventMap pending myTickEvent tickTime syncList idleEvent schedulerProcess suspended myScripts'
	classVariableNames: 'Debug'
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptScheduler class methodsFor: 'instance creation' stamp: 'ar 1/30/2005 11:57'!
new
	^self basicNew initialize.! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 7/14/2001 00:03'!
activeList
	^activeList! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 7/14/2001 00:04'!
activeScript
	^activeScript! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 6/11/2002 19:38'!
addLink: link toList: aList
	link nextLink: aList.
	aList ifNotNil:[aList prevLink: link].
	^link! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 7/28/2010 09:00'!
bootstrap: aBlock
	"Run aBlock in the context of this scheduler"
	| script |
	script := aBlock asAsyncScriptMessageIn: self arguments: #().
	script setScheduler: self.
	script := script valueWithEvent: ScriptEvent new.
	[script isDone] whileFalse:[self run].! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 2/18/2002 00:54'!
debugPrintActive
	"ActiveHand lastEvent shiftPressed ifTrue:[Transcript cr; show: activeScript]."! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/14/2001 00:04'!
freePendingScripts
	"Free any pending scripts that have been asynchronously released"
	interruptProtect critical:[
		scheduledList := self removeLinksWithValue: nil from: scheduledList.
		pending := false].! !

!ScriptScheduler methodsFor: 'testing' stamp: 'ar 5/24/2005 22:06'!
hasActiveScripts
	"Answer whether the scheduler has any active scripts which need to be run"
	^activeList isEmpty not! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 1/8/2008 17:47'!
initialize
	activeList := Semaphore new.
	syncList := Semaphore new.
	scheduledList := nil.
	pending := false.
	suspended := false.
	interruptProtect := Semaphore forMutualExclusion.
	myTickEvent := ScriptTickEvent new setSender: self selector: #tick arguments: {0. 0}.
	myEventMap := nil.
	tickTime := Time millisecondClockValue.
	tickList := Semaphore new.
! !

!ScriptScheduler methodsFor: 'testing' stamp: 'ar 7/14/2001 00:04'!
isEmpty
	^activeList isEmpty and:[scheduledList == nil]! !

!ScriptScheduler methodsFor: 'testing' stamp: 'ar 7/14/2001 00:04'!
isScriptActive: aScript
	^aScript suspendingList == activeList! !

!ScriptScheduler methodsFor: 'testing' stamp: 'ar 8/21/2001 23:33'!
isScriptScheduler
	"Yes. I am."
	^true! !

!ScriptScheduler methodsFor: 'event handling' stamp: 'ar 7/14/2001 00:05'!
myEventMap
	^myEventMap! !

!ScriptScheduler methodsFor: 'event handling' stamp: 'ar 7/14/2001 00:05'!
myEventMap: aDictionary
	myEventMap := aDictionary! !

!ScriptScheduler methodsFor: 'private' stamp: 'ar 5/22/2005 22:35'!
myScripts
	^myScripts ifNil:[#()]! !

!ScriptScheduler methodsFor: 'private' stamp: 'ar 5/22/2005 22:35'!
myScripts: scriptList
	myScripts := scriptList.
! !

!ScriptScheduler methodsFor: 'accessing' stamp: ''!
myTickEvent
	^myTickEvent! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 10/26/2007 15:21'!
proceedActiveScripts
	"Proceed running all active scripts in the current time slot.
	NOTE: This is a hack we currently need in order to deal with layout changes accordingly."
	"Prevent recursive calls to the scheduler.
	We need this since there are still callers of #doOneCycle"
	Processor activeProcess == activeScript ifTrue:[^self].
	"If any scripts need to be released, release them now"
	pending ifTrue:[self freePendingScripts].

	["Process any pending scripts that may have started in the mean time"
	[(activeScript := activeList firstOrNil) == nil]
		whileFalse:[self debugPrintActive. activeList signal].
	"Process synchronously waiting scripts in REVERSE order"
	syncList isEmpty] whileFalse:[syncList last transferTo: activeList].
! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 6/11/2002 19:40'!
removeLink: aLink fromList: aList
	| newHead |
	newHead := aList.
	aLink prevLink == nil
		ifTrue:[newHead := aLink nextLink]
		ifFalse:[newHead := aList.
				aLink prevLink nextLink: aLink nextLink].
	aLink nextLink == nil
		ifFalse:[aLink nextLink prevLink: aLink prevLink].
	^newHead! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 6/11/2002 19:40'!
removeLinkSuchThat: aBlock from: aList
	| newHead next prev |
	newHead := next := aList.
	prev := nil.
	[next == nil] whileFalse:[
		(aBlock value: next) 
			ifTrue:[next := next nextLink]
			ifFalse:[prev := next. next := nil]].
	newHead := prev ifNil:[^nil].
	next := prev nextLink.
	[next == nil] whileFalse:[
		(aBlock value: next) ifTrue:[
			prev nextLink: next nextLink.
			prev nextLink ifNotNil:[prev nextLink prevLink: prev].
			next := next nextLink.
		] ifFalse:[
			prev := next.
			next := next nextLink.
		].
	].
	^newHead! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 6/11/2002 19:39'!
removeLinksWithValue: value from: aList
	"Remove all links with the given value from aList. 
	Return the new head of the list"
	| newHead next prev |
	newHead := next := aList.
	prev := nil.
	[next == nil] whileFalse:[
		(next value == value) 
			ifTrue:[next := next nextLink]
			ifFalse:[prev := next. next := nil]].
	newHead := prev ifNil:[^nil].
	next := prev nextLink.
	[next == nil] whileFalse:[
		(next value == value) ifTrue:[
			prev nextLink: next nextLink.
			prev nextLink ifNotNil:[prev nextLink prevLink: prev].
			next := next nextLink.
		] ifFalse:[
			prev := next.
			next := next nextLink.
		].
	].
	^newHead! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 4/24/2002 15:06'!
removeLinksWithValue: value from: aList andDo: aBlock
	"Remove all links with the given value from aList. 
	Return the new head of the list"
	| newHead next prev val |
	newHead := next := aList.
	prev := nil.
	[next == nil] whileFalse:[
		((val := next value) == value) 
			ifTrue:[next := next nextLink]
			ifFalse:[prev := next. next := nil. aBlock value: val]].
	newHead := prev ifNil:[^nil].
	next := prev nextLink.
	[next == nil] whileFalse:[
		((val := next value) == value) ifTrue:[
			prev nextLink: next nextLink.
			prev nextLink ifNotNil:[prev nextLink prevLink: prev].
			next := next nextLink.
		] ifFalse:[
			prev := next.
			next := next nextLink.
			aBlock value: val.
		].
	].
	^newHead! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 5/22/2005 17:14'!
resume
	"Suspend this scheduler"
	self suspended ifFalse:[^self].
	suspended := false.
	schedulerProcess ifNotNil:[schedulerProcess resume]
! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/14/2001 00:04'!
resumeScript: aUserScript
	"Resume the given user script"
	aUserScript == Processor activeProcess 
		ifTrue:[^self error:'Can only resume inactive scripts'].
	aUserScript setScheduler: self.
	aUserScript resume.! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 10/26/2007 15:21'!
run
	"Run all active scripts in the current time slot. An interesting bit of semantics here is that if the scheduler is run from within *another* scripting process its behavior is considered atomic. E.g., *no* script inside the caller will be run until we're done here."
	| priority process worker |
	"Prevent recursive calls to the scheduler."
	(process := Processor activeProcess) == activeScript 
		ifTrue:[^self error: 'Cannot drive myself'].

	process isScriptingProcess ifTrue:[
		"This process was driven by another scheduler. 
		To be atomic, we suspend that scheduler."
		process scheduler suspend. "g'night"

		"Also suspend the process itself and fork off a non-scripting
		process to drive the scheduler."
		worker := [
			self run.
			process scheduler resume. "g'mornin"
			process resume.
		] forkAt: Processor userSchedulingPriority.
		^[process suspend] ensure:[
			worker terminate.
			schedulerProcess := nil.
		].
	].
	schedulerProcess := Processor activeProcess.
	priority := schedulerProcess priority.
	schedulerProcess priority: Processor userSchedulingPriority.

	"If any scripts need to be released, release them now"
	pending ifTrue:[self freePendingScripts].

	"Process pending active scripts"
	[
		[(activeScript := activeList firstOrNil) == nil]
			whileFalse:[self debugPrintActive. activeList signal].

	"Process synchronously waiting scripts in REVERSE order"
	syncList isEmpty] whileFalse:[syncList last transferTo: activeList].

	schedulerProcess ifNotNil:[schedulerProcess priority: priority].
	schedulerProcess := nil.! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 10/26/2007 15:21'!
runActiveScripts
	"Run all active scripts in the current time slot. An interesting bit of semantics here is that if the scheduler is run from within *another* scripting process its behavior is considered atomic. E.g., *no* script inside the caller will be run until we're done here."
	| toTick priority process worker |
	"Prevent recursive calls to the scheduler.
	We need this since there are still callers of #doOneCycle"
	(process := Processor activeProcess) == activeScript ifTrue:[^self].

	process isScriptingProcess ifTrue:[
		"This process was driven by another scheduler. 
		To be atomic, we suspend that scheduler."
		process scheduler suspend. "g'night"

		"Also suspend the process itself and fork off a non-scripting
		process to drive the scheduler."
		worker := [
			self runActiveScripts.
			process scheduler resume. "g'mornin"
			process resume.
		] forkAt: Processor userSchedulingPriority.
		^[process suspend] ensure:[
			worker terminate.
			schedulerProcess := nil.
		].
	].
	schedulerProcess := Processor activeProcess.
	priority := schedulerProcess priority.
	schedulerProcess priority: Processor userSchedulingPriority.

	"If any scripts need to be released, release them now"
	pending ifTrue:[self freePendingScripts].

	"Prepare for the next cycle of processing scripts"
	tickTime := Time millisecondClockValue.
	myTickEvent setArguments: (Array with: tickTime).
	toTick := tickList.
	tickList := Semaphore new.

	"Process pending active scripts first. This is so the 'main' UI process can spawn the redraw and the event processing script which should be processed before any tickers."
	[(activeScript := activeList firstOrNil) == nil]
		whileFalse:[self debugPrintActive. activeList signal].

	"Process all intrinsic tickers; this is equivalent to sending the #tick for objects registered with the tick list."
	[(activeScript := toTick firstOrNil) == nil] 
		whileFalse:[self debugPrintActive. toTick signal].

	"But also signal #tick explicitly in case we have some additional dependents"
	self signalEvent: myTickEvent clone.
	["Process any scripts that may have started in the mean time"
	[(activeScript := activeList firstOrNil) == nil]
		whileFalse:[self debugPrintActive. activeList signal].

	"Process synchronously waiting scripts in REVERSE order"
	syncList isEmpty] whileFalse:[syncList last transferTo: activeList].

	schedulerProcess priority: priority.
	schedulerProcess := nil.! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 4/23/2002 21:16'!
scheduleScript: aUserScript
	| link |
	aUserScript == Processor activeProcess
		ifFalse:[^self error:'Can only schedule the active script'].
	aUserScript isScheduled
		ifTrue:[self error:'This script is already scheduled'].
	aUserScript == activeScript ifFalse:[activeList primitiveWait].
	"now we're in sync with the scheduler"
	link := ScriptProcessLink weakOn: aUserScript.
	scheduledList := self addLink: link toList: scheduledList.
	self signal: #scheduleScript with: aUserScript.
	^link! !

!ScriptScheduler methodsFor: 'accessing' stamp: ''!
scheduledScripts
	^Array streamContents:[:s|
		self scheduledScriptsDo:[:p| s nextPut: p].
	].! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/20/2001 22:00'!
scheduledScriptsDo: aBlock
	"Evaluate aBlock for all scripts scheduled in the receiver."
	| link |
	link := scheduledList.
	[link == nil] whileFalse:[
		link process == nil ifFalse:[aBlock value: link process].
		link := link nextLink.
	].! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 4/26/2005 16:44'!
schedulerProcess
	"Answer the process currently driving this scheduler"
	^schedulerProcess! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/17/2001 17:54'!
scriptsDo: aBlock
	"Evaluate aBlock for all scripts associated with the receiver.
	NOTE: Some scripts may never have been activated before.
	Because of this they are not (yet) in the scheduled list but
	waiting on the active list to be scheduled. To enumerate these
	scripts as well we need to check the processes waiting in the
	active list for any that have not been started yet."
	| link |
	link := scheduledList.
	[link == nil] whileFalse:[
		link process == nil ifFalse:[aBlock value: link process].
		link := link nextLink.
	].
	link := activeList firstOrNil.
	[link == nil] whileFalse:[
		link isStarted ifFalse:[aBlock value: link].
		link := link nextLink].! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 10/26/2007 15:20'!
shutDown
	"Shut down the receiver"
	| aScript |
	self scheduledScriptsDo:[:script| script terminate].
	"Terminate activeList explicitly - we may have scripts waiting to be scheduled there"
	[aScript := activeList firstOrNil.
	aScript == nil] whileFalse:[aScript terminate].! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 5/22/2005 16:33'!
suspend
	"Suspend this scheduler"
	schedulerProcess ifNil:[^self].
	schedulerProcess isSuspended ifTrue:[^self].
	schedulerProcess suspend.
	suspended := true.
! !

!ScriptScheduler methodsFor: 'testing' stamp: 'ar 5/22/2005 17:16'!
suspended
	^suspended == true! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 2/13/2002 18:16'!
syncList
	"Return the list used for serializing asynchronous scripts. This list is used for scripts triggering async events that need to be responded to synchronously."
	^syncList ifNil:[syncList := Semaphore new].! !

!ScriptScheduler methodsFor: 'initialize' stamp: 'ar 5/22/2005 16:33'!
terminate
	"Suspend this scheduler"
	schedulerProcess ifNotNil:[
		schedulerProcess terminate.
		schedulerProcess := nil.
	].
! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/14/2001 00:05'!
terminateScript: aUserScript
	aUserScript terminate.! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 3/14/2003 04:29'!
tickEvent
	^myTickEvent clone! !

!ScriptScheduler methodsFor: 'accessing' stamp: 'ar 7/20/2001 13:53'!
tickingList
	"Return the list into which objects can put themselves to wait for a single tick. This list is exposed because deep call chains can be very inefficient when waiting for many ticks. The ticking list should be used primitively, e.g., by calling 'self clock tickingList primitiveWait'."
	^tickList! !

!ScriptScheduler methodsFor: 'scheduling' stamp: 'ar 7/22/2001 00:56'!
unscheduleScript: aLink
	"May be called asynchronously, e.g., when explicitly terminating scripts
	from debugger etc. Need to make sure we don't interfere with
	scheduledLists activity"
	self signal: #unscheduleScript with: aLink process.
	interruptProtect critical:[
		pending := true.
		aLink process: nil.
	].! !

!Array methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 15:42'!
asEventSpecificationIn: owner
	self size = 0 ifTrue:[^nil].
	^self! !

!BlockClosure methodsFor: '*Scripting-Extensions' stamp: 'eem 8/24/2008 10:43'!
asAsyncScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	^self scriptEvaluationSelector
		ifNil: [super asAsyncScriptMessageIn: scriptPerformer arguments: args]
		ifNotNil:
			[:selector|
			Processor activeIsland asyncScriptMessageSend 
				receiver: scriptPerformer 
				selector: selector
				arguments: {self},args]! !

!BlockClosure methodsFor: '*Scripting-Extensions' stamp: 'eem 8/24/2008 10:42'!
asScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	^self scriptEvaluationSelector
		ifNil: [super asAsyncScriptMessageIn: scriptPerformer arguments: args]
		ifNotNil:
			[:selector|
			ScriptMessageSend 
				receiver: scriptPerformer 
				selector: selector
				arguments: {self},args]! !

!BlockClosure methodsFor: '*Scripting-Extensions' stamp: 'eem 8/24/2008 10:41'!
isScriptDescriptorIn: aScript
	"Answer if the receiver is equivalent to the argument script, a ScriptMessageSend"
	^aScript selector == self scriptEvaluationSelector
	  and: [(aScript arguments at: 1) == self]! !

!BlockClosure methodsFor: '*Scripting-Extensions' stamp: 'eem 8/24/2008 10:46'!
scriptEvaluationSelector
	"Answer the right selector to evaluate the receiver when converted to a script, or nil if there isn't one."
	^numArgs <= 4 ifTrue:
		[#(	evaluate:
			evaluate:with:
			evaluate:with:with:
			evaluate:with:with:with:
			evaluate:with:with:with:with:) at: numArgs + 1]! !

!Symbol methodsFor: '*Scripting-Extensions' stamp: 'ar 3/26/2007 22:44'!
asAsyncScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	^Processor activeIsland asyncScriptMessageSend 
		receiver: scriptPerformer selector: self arguments: args! !

!Symbol methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 15:41'!
asEventSpecificationIn: owner
	^Array with: owner with: self! !

!Symbol methodsFor: '*Scripting-Extensions' stamp: 'ar 7/24/2001 21:18'!
asScriptMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	^ScriptMessageSend receiver: scriptPerformer selector: self arguments: args! !

!Symbol methodsFor: '*Scripting-Extensions' stamp: 'ar 2/6/2004 17:46'!
asTickMessageIn: scriptPerformer arguments: args
	"Convert the receiver into an appropriate ScriptMessageSend"
	^TickMessageSend receiver: scriptPerformer selector: self arguments: args! !

!Symbol methodsFor: '*Scripting-Extensions' stamp: 'ar 4/24/2002 14:49'!
isScriptDescriptorIn: aScript
	^aScript selector == self! !

!MessageSend methodsFor: '*Scripting-Extensions' stamp: 'ar 5/18/2002 17:18'!
printMessage
	^String streamContents:[:s| self printMessageOn: s].! !

!MessageSend methodsFor: '*Scripting-Extensions' stamp: 'ar 5/18/2002 17:50'!
printMessageOn: aStream
	aStream print: receiver identityHash.
	aStream nextPutAll: (receiver printStringLimitedTo: 50).
	aStream nextPutAll:'>>'.
	aStream nextPutAll: (selector printStringLimitedTo: 50).
	aStream print: arguments.! !

MessageSend subclass: #ScriptMessageSend
	instanceVariableNames: 'eventSpec extraArgs paused myFarRef'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

ScriptMessageSend subclass: #AsyncScriptMessageSend
	instanceVariableNames: 'myProcess myQueue myIsland myScheduler myTrigger myScripts accessLock myEventMap'
	classVariableNames: 'UpdateLock'
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!AsyncScriptMessageSend class methodsFor: 'class initialization' stamp: 'ar 7/27/2007 15:56'!
initialize
	"AsyncScriptMessageSend initialize"
	UpdateLock := Semaphore forMutualExclusion.
! !

!AsyncScriptMessageSend class methodsFor: 'instance creation' stamp: 'ar 6/11/2002 19:32'!
receiver: anObject selector: aSymbol arguments: anArray
	^ self new
		receiver: anObject;
		selector: aSymbol;
		arguments: anArray;
		setDefaultScheduler! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 7/27/2007 15:53'!
accessLock
	^accessLock ifNil:[UpdateLock critical:[
		accessLock ifNil:[accessLock := Semaphore forMutualExclusion]]].
! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 5/7/2005 18:26'!
autoStartTrigger
	myTrigger ifNil:[^nil].
	myTrigger changeEvent ifNil:[^nil].
	self startScript: #onEventSourceChanged when: {receiver. myTrigger changeEvent}.
	self onEventSourceChanged.! !

!AsyncScriptMessageSend methodsFor: 'testing' stamp: ''!
isRunning
	^myProcess notNil! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'bgf 9/12/2008 10:40'!
myEventMap
	^ myEventMap! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'bgf 9/12/2008 10:41'!
myEventMap: map
	myEventMap := map! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 8/5/2003 20:56'!
myScripts
	^myScripts ifNil:[#()]! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 8/5/2003 20:56'!
myScripts: newScripts
	myScripts := newScripts! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 2/7/2004 16:19'!
myTrigger
	^myTrigger! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 2/7/2004 16:20'!
myTrigger: aTrigger
	myTrigger := aTrigger.! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 2/28/2005 15:49'!
onEventSourceChanged
	"My event source might have changed"
	| oldSource newSource |
	eventSpec ifNil:[^self].
	myTrigger ifNil:[^self].
	oldSource := eventSpec at: 1.
	newSource := [receiver getField: myTrigger field] on: Error do:[:ex| ex return: nil].
	newSource == oldSource ifFalse:[self onUpdateTrigger: newSource from: oldSource].! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 5/7/2005 18:39'!
onUpdateTrigger: newValue from: oldValue
	eventSpec ifNil:[^self].
	"This is pretty ugly ... but we really need to stop any message which may have been scheduled based on the old event spec"
	myQueue := nil.
	myProcess ifNotNil:[myProcess terminate].

	(eventSpec at: 1) on: (eventSpec at: 2) unNotify: self.
	eventSpec at: 1 put: newValue.
	self isPaused ifFalse:[(eventSpec at: 1) on: (eventSpec at: 2) notify: self].! !

!AsyncScriptMessageSend methodsFor: 'accessing' stamp: 'ar 1/22/2005 21:14'!
pauseScript
	"Stop notifications from event sources"
	super pauseScript.
	myQueue := nil.! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'jh 5/5/2003 19:11'!
requestStop
	myProcess ifNotNil:[myProcess terminate]! !

!AsyncScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 4/24/2002 14:48'!
runningScript
	^myProcess! !

!AsyncScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 1/8/2008 17:42'!
scriptStopped: aScript
	aScript == myProcess ifTrue:[
		myProcess := nil.
		myQueue := nil.
		eventSpec ifNil:[self terminateScript].
	].! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 5/22/2005 21:02'!
setDefaultScheduler
	myIsland := Processor activeIsland.
	myScheduler := myIsland scheduler ifNil:[Processor activeProcess scheduler].! !

!AsyncScriptMessageSend methodsFor: 'private' stamp: 'ar 5/22/2005 16:29'!
setScheduler: aScheduler
	myScheduler := aScheduler! !

!AsyncScriptMessageSend methodsFor: 'accessing' stamp: 'ar 5/19/2002 23:48'!
terminateScript
	Processor activeProcess == myProcess 
		ifTrue:[^self startScript: #terminateScript].
	super terminateScript.
	myProcess ifNotNil:[
		myProcess terminate.
		myProcess := nil].! !

!AsyncScriptMessageSend methodsFor: 'evaluating' stamp: 'bgf 2/11/2008 12:17'!
value
	| event eventList result coalesce |
	myQueue ifNil:[^nil].
	[self accessLock critical:[
		myQueue ifNil:[^nil].
		eventList := nil.
		"pull out all the events from the queue that can be coalesced and chain them 
		in reverse order to have the newest event at the head of the list"
		(event := myQueue firstOrNil) ifNotNil:[
			eventList := ScriptEventList new.
			eventList addFirst: myQueue removeFirst.
			coalesce := event coalesceEvents.
			coalesce ifTrue:["transfer all the events"
				[myQueue isEmpty] whileFalse:[eventList addFirst: myQueue removeFirst].
			].
		].
	].
	eventList == nil] whileFalse:[
		event := eventList first.
		result := [self synchronousValueWithEvent: event] 
			on: DropAllButLastEvent do:[:ex|
				myQueue isEmpty ifTrue:[ex resume]. "with old event"
				[myQueue isEmpty] whileFalse:[event := myQueue removeFirst].
				ex retry. "with new event"
		].
		coalesce ifTrue:[^result].
	].
	^nil! !

!AsyncScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 6/15/2009 13:46'!
valueWithEvent: anEvent

	"Serialize access since different processes may get here at the same time"
	self accessLock critical:[ | myScript proc |
		myQueue ifNil:[myQueue := ScriptEventList new].
		myQueue addLast: anEvent.
		myProcess ifNotNil:[^self].
		proc := (ScriptProcess forIsland: myIsland) newScript.
		proc message: self.
		proc owner: self.
		proc setScheduler: myScheduler.
		proc event: anEvent.
		(myScript := Processor activeProcess) traceEvents ifTrue:[
			myScript traceEvent: anEvent message: self script: proc.
			proc eventTracer: myScript eventTracer.
			proc traceEvents: true.
		].
		proc hand: myScript hand.
		myProcess := proc.
		myProcess primitiveResume. "needs to be primitive to make this and prev statement atomic"
	].
	^myProcess
! !

!ScriptMessageSend methodsFor: 'initialize' stamp: 'ar 5/7/2005 18:24'!
autoStartTrigger
	"If my trigger needs special startup do it now"! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 4/28/2002 00:50'!
computeExtraArgs
	^extraArgs ifNil:[
		extraArgs := selector numArgs "how many do we need?"
					- self arguments size "how many do we have?".
		"resize arguments to fit"
		extraArgs < 0 ifTrue:[
			arguments := arguments copyFrom: 1 to: arguments size + extraArgs.
			extraArgs := 0].
		extraArgs > 0 ifTrue:[arguments := arguments, (Array new: extraArgs)].
		extraArgs].! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 4/28/2002 16:32'!
eventSpec
	^eventSpec ifNil:[#()].! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 4/24/2002 15:43'!
eventSpec: anArray
	eventSpec := anArray! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 4/28/2002 00:49'!
extraArgs
	^extraArgs ifNil:[self computeExtraArgs].! !

!ScriptMessageSend methodsFor: 'testing' stamp: 'ar 2/6/2004 17:45'!
isEventTriggered
	"Am I triggered by some event?"
	^eventSpec notNil and:[eventSpec size >= 2]! !

!ScriptMessageSend methodsFor: 'testing' stamp: 'ar 9/23/2003 01:18'!
isPaused
	^paused == true! !

!ScriptMessageSend methodsFor: 'testing' stamp: 'ar 4/24/2002 17:02'!
isScriptDescriptorIn: aScript
	^aScript == self! !

!ScriptMessageSend methodsFor: 'testing' stamp: 'ar 3/23/2003 21:34'!
isTickingScript
	^false! !

!ScriptMessageSend methodsFor: 'testing' stamp: 'ar 9/17/2008 15:50'!
isTriggeredBy: signaler
	"Answer true if this script is triggered by an event from the given signaler"
	eventSpec ifNil:[^false].
	1 to: eventSpec size by: 2 do:[:i| (eventSpec at: i) == signaler ifTrue:[^true]].
	^false! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 5/22/2005 16:21'!
myFarRef
	^myFarRef! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 10/23/2006 19:01'!
myFarRef: farRef
	"Hang on to my far ref since the far ref might only be referenced weakly by a (far) event source and loosing the far ref means no events will reach us."
	myFarRef ifNotNil:[
		"This should really NEVER happen but it currently does due to a major bug in the island implementation. This needs fixing but for the interim period let's at least tell people that things are going wrong..."
		WorldState addDeferredUIMessage: (MessageSend receiver: Transcript selector: #show: argument: '\WARNING: Overriding farRef in ' withCRs, self printString).
	].
	myFarRef := farRef! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 9/23/2003 01:14'!
myTrigger
	^nil! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 4/24/2002 12:21'!
noticeEvent: anEvent
	selector ifNotNil:[^self valueWithEvent: anEvent]! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 5/12/2005 20:53'!
pauseScript
	"Stop notifications from event sources"
	self isPaused ifTrue:[^self].
	eventSpec ifNil:[^self].
	1 to: eventSpec size by: 2 do:[:i|
		(eventSpec at: i) on: (eventSpec at: i+1) unNotify: self.
	].
	paused := true.
	receiver signal: #scriptStatusChanged with: selector.! !

!ScriptMessageSend methodsFor: 'factory' stamp: 'ar 5/16/2002 22:11'!
readArgumentsFrom: anArray
	arguments := Array new: anArray size - 1.
	2 to: anArray size do:[:idx|
		arguments at: idx-1 put: (anArray at: idx).
	].! !

!ScriptMessageSend methodsFor: 'factory' stamp: 'ar 5/16/2002 22:10'!
readEventsFrom: anArray
	| index key value events rcvr sel evtSpec spec |
	events := Array new: anArray size - 1 * 2.
	index := 0.
	2 to: anArray size do:[:idx|
		spec := anArray at: idx.
		key := spec first.
		key == #event ifFalse:[self error:'#event expected'].
		rcvr := sel := nil.
		2 to: spec size do:[:i|
			evtSpec := spec at: i.
			key := evtSpec first.
			value := evtSpec last.
			key == #receiver ifTrue:[rcvr := value].
			key == #selector ifTrue:[sel := value].
		].
		events at: (index := index + 1) put: rcvr.
		events at: (index := index + 1) put: sel.
	].
	eventSpec := events! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 5/12/2005 20:54'!
resumeScript
	"Resume notifications from event sources"
	self isPaused ifFalse:[^self].
	eventSpec ifNil:[^self].
	1 to: eventSpec size by: 2 do:[:i|
		(eventSpec at: i) on: (eventSpec at: i+1) notify: self.
	].
	paused := false.
	receiver signal: #scriptStatusChanged with: selector.! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 4/24/2002 14:48'!
runningScript
	^nil! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 11/12/2004 22:02'!
synchronousValueWithArguments: anArray event: anEvent
	"WARNING: The debugger knows about the existence of this method!!"
	| result myScript |
	(myScript := Processor activeProcess) traceEvents 
		ifTrue:[myScript traceEvent: anEvent message: self script: nil].
	result := receiver perform: selector withArguments: anArray.
	eventSpec ifNil:[receiver privateRemoveScript: self].
	^result! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 4/21/2005 22:14'!
synchronousValueWithEvent: anEvent
	| args count index |
	extraArgs ifNil:[self computeExtraArgs]. "resize"
	(extraArgs = 0 "no extra args needed"
		or:[anEvent == nil]) "or no extra args available"
			ifTrue:[^self synchronousValueWithArguments: arguments event: anEvent].
	"Need some arguments from args"
	args := arguments clone.
	index := arguments size - extraArgs.
	count := extraArgs min: anEvent arguments size.
	args replaceFrom: index+1 to: index+count with: anEvent messageArguments startingAt: 1.
	index := index + count.
	index < args size ifTrue:[args at: index+1 put: anEvent].
	^self synchronousValueWithArguments: args event: anEvent! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 5/12/2005 20:54'!
terminateScript
	eventSpec ifNotNil:[
		1 to: eventSpec size by: 2 do:[:i|
			(eventSpec at: i) on: (eventSpec at: i+1) unNotify: self.
		].
		eventSpec := nil].
	receiver privateRemoveScript: self.
	receiver signal: #scriptStatusChanged with: selector.! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 8/22/2001 01:01'!
value
	^self value: #()! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 4/24/2002 12:33'!
value: anEvent
	^self valueWithEvent: anEvent! !

!ScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 3/30/2003 01:30'!
valueWithEvent: anEvent
	^self synchronousValueWithEvent: anEvent
! !

!ScriptMessageSend methodsFor: 'accessing' stamp: 'ar 2/24/2004 14:15'!
wakeupScript
	"Resume notifications from event sources"
	eventSpec ifNil:[^self].
	1 to: eventSpec size by: 2 do:[:i|
		(eventSpec at: i) on: (eventSpec at: i+1) notify: self.
	].
	paused := false.! !

ScriptMessageSend subclass: #TScriptMessageSend
	instanceVariableNames: 'myScripts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!TScriptMessageSend commentStamp: '<historical>' prior: 0!
An asynchronous script message to be used with Croquet (i.e., utilizing #future instead of script processes).!

!TScriptMessageSend methodsFor: 'private' stamp: 'ar 3/26/2007 22:42'!
myScripts
	^myScripts ifNil:[#()]! !

!TScriptMessageSend methodsFor: 'private' stamp: 'ar 3/26/2007 22:42'!
myScripts: newScripts
	myScripts := newScripts! !

!TScriptMessageSend methodsFor: 'evaluating' stamp: 'ar 3/26/2007 22:57'!
valueWithEvent: anEvent
	self future synchronousValueWithEvent: anEvent.! !

ScriptMessageSend subclass: #TickMessageSend
	instanceVariableNames: 'frequency myProcess myTrigger'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/15/2003 18:48'!
frequency
	^frequency! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/15/2003 18:48'!
frequency: aFrequency
	frequency := aFrequency.! !

!TickMessageSend methodsFor: 'testing' stamp: 'ar 3/23/2003 21:33'!
isTickingScript
	^true! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/15/2003 18:56'!
myProcess
	^myProcess! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/15/2003 18:57'!
myProcess: aProcess
	myProcess := aProcess.! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 8/15/2003 01:42'!
myTrigger
	^myTrigger! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 8/15/2003 01:42'!
myTrigger: aTrigger
	myTrigger := aTrigger! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/23/2003 21:55'!
pauseScript
	myProcess ifNotNil:[
		myProcess terminateScript.
		myProcess := nil.
	].
	super pauseScript.! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 5/17/2005 11:41'!
resumeScript
	super resumeScript.
	myProcess ifNil:[
		myProcess := receiver startScript: #runTicking:at: withArguments:{self. frequency}.
		myProcess := myProcess message.
	].
	receiver signal: #scriptStatusChanged with: selector.! !

!TickMessageSend methodsFor: 'accessing' stamp: 'ar 3/15/2003 18:47'!
terminateScript
	self pauseScript.
	super terminateScript.! !

!LinkedList methodsFor: '*Scripting-Extensions' stamp: 'ar 7/8/2001 14:24'!
firstOrNil
	"Answer the first element of the receiver or nil if empty"
	^firstLink! !

!LinkedList methodsFor: '*Scripting-Extensions' stamp: 'ar 3/5/2002 23:17'!
includes: aLink
	| link |
	link := firstLink.
	[link == nil] whileFalse:[
		aLink == link ifTrue:[^true].
		link := link nextLink].
	^false! !

LinkedList subclass: #ScriptEventList
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ProcessorScheduler class methodsFor: '*Scripting-override' stamp: 'ar 7/8/2001 16:39'!
initialize
     
	SystemRockBottomPriority := 10.
	SystemBackgroundPriority := 20.
	UserBackgroundPriority := 30.
	UserSchedulingPriority := 40.
	UserInterruptPriority := 50.
	LowIOPriority := 60.
	HighIOPriority := 70.
	TimingPriority := 80.

	"ProcessorScheduler initialize."! !

!ProcessorScheduler methodsFor: '*Scripting-Extensions' stamp: 'ar 3/7/2006 13:08'!
initialScriptPriority
	"Return the priority at which a user script process should be started.
	NOTE: The initial script priority needs to be higher than the priority
	during execution of the script. This is so that any scripts which are scheduled
	from another script preempt the running script to get in sync with the scheduler."
	^self userScriptPriority + 1! !

!ProcessorScheduler methodsFor: '*Scripting-Extensions' stamp: 'ar 8/21/2001 23:33'!
isScriptScheduler
	"No, I am not. I schedule 'regular' processes."
	^false! !

!ProcessorScheduler methodsFor: '*Scripting-Extensions' stamp: 'ar 3/7/2006 13:08'!
userScriptPriority
	"Return the priority at which a user script process should be run"
	^45! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 11/10/2004 17:57'!
clock
	"To prevent infinite recursion"
	^nil! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2002 20:16'!
eventTracer
	^nil! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2002 20:17'!
eventTracer: aTracer
	"Ignored"! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 7/7/2001 15:30'!
isScriptingProcess
	^false! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 3/4/2002 12:42'!
isSystemProcess
	^true! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 12/8/2009 20:13'!
island
	^island! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 12/8/2009 21:56'!
island: anIsland
	island := anIsland! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 11/10/2004 17:53'!
scheduler
	^nil! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2002 20:12'!
traceEvents
	^false! !

!Process methodsFor: '*Scripting-Extensions' stamp: 'ar 5/17/2002 20:16'!
traceEvents: aBool
	"ignored"! !

Process subclass: #ScriptProcess
	instanceVariableNames: 'owner event message result flags properties scheduler myEventMap myLink myTriggerBlock myHandler topContext'
	classVariableNames: 'FlagScriptDone FlagScriptSignalResume FlagScriptSignalWait FlagScriptStarted FlagScriptStopRequest FlagScriptTraceEvents'
	poolDictionaries: ''
	category: 'Scripting-Kernel'!

!ScriptProcess class methodsFor: 'class initialization' stamp: 'ar 12/10/2009 00:53'!
initialize
	"ScriptProcess initialize"
	FlagScriptStarted := 1.
	FlagScriptDone := 2.
	FlagScriptSignalWait := 4.
	FlagScriptSignalResume := 8.
	FlagScriptTraceEvents := 16.
	FlagScriptStopRequest := 32.
	
	(Process instVarNames includes: 'island') 
		ifFalse:[[Process addInstVarName: 'island'] on: Warning do:[:ex| ex resume]].
! !

!ScriptProcess class methodsFor: 'instance creation' stamp: 'ar 4/24/2002 14:16'!
newScript
	^self new newScript! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 6/10/2002 20:15'!
beginCritical: aSemaphore
	self beginWait.! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 5/22/2005 16:23'!
beginWait
	"Signal the beginning of a wait cycle"
	| time started |
	"Boost this scripts priority for the wait. This is necessary in order to keep both, scripts newly triggered (forked) by an event and those waiting for some event in the right order (e.g., without the boost a waiting script may run :=after:= one that was 'created later' based on the same event)."
	priority := priority + 1.

	(flags bitAnd: FlagScriptSignalWait) = 0 ifFalse:[
		time := Time millisecondClockValue.
		(flags bitAnd: FlagScriptSignalResume) = 0 
			ifTrue:[started := time]
			ifFalse:[started := self valueOfProperty: #startTime ifAbsent:[time]].
		self signal: #waiting with: time with: (time - started)].
! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 6/10/2002 20:16'!
beginWaitOn: aSemaphore
	self beginWait.
! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/9/2001 13:52'!
clock
	^scheduler! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 11/12/2004 16:34'!
displayText
	^(String streamContents:[:s|
		"s nextPutAll: self event selector; space."
false ifTrue:[
		s nextPutAll: (self event sender printString contractTo: 50).
		s nextPutAll:'>>'.
		s nextPutAll: (self event selector).
] ifFalse:[
		s nextPutAll: (self owner receiver class printString contractTo: 50).
		s nextPutAll:'>>'.
		s nextPutAll: (self messageName ifNil:['???']).
].
	]).! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 9/25/2009 12:13'!
endCritical: aSemaphore
	"Resume after leaving a critical section"
	self isDone ifTrue:[^self]. "don't resume during unwind"
	scheduler ifNil:[^self].
	(self isActiveScript)
		ifFalse:[scheduler activeList primitiveWait].
	self endWait.! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 5/22/2005 16:23'!
endWait
	"Signal the ending of a wait cycle"
	| time |
	priority := priority - 1. "restore previous priority"
	(flags bitAnd: FlagScriptSignalResume) = 0 ifFalse:[
		time := Time millisecondClockValue.
		(flags bitAnd: FlagScriptSignalWait) = 0 ifFalse:[self setProperty: #startTime toValue: time].
		self signal: #resuming with: time].! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 6/10/2002 20:16'!
endWaitOn: aSemaphore
	"Resynchronize the receiver with the scheduler."
	(self isActiveScript)
		ifFalse:[scheduler activeList primitiveWait].
	"Note: The reason for not resynchronizing when we're the active script is so that certain 'atomic' notions can be preserved. E.g., when we enter a signaled semaphore, the script should :=not:= be rescheduled at the end of the scheduler since the meaning of entering a signaled semaphore is non-preemptive."
	self endWait.
	^nil! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 11/12/2004 13:04'!
event
	^event! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 11/12/2004 13:04'!
event: anEvent
	event := anEvent! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 7/20/2001 18:56'!
eventHandlerFor: anEvent
	^self nextEventHandlerFor: anEvent sender selector: anEvent selector! !

!ScriptProcess methodsFor: 'event trace' stamp: 'ar 5/17/2002 20:18'!
eventTracer
	^self valueOfProperty: #eventTracer
! !

!ScriptProcess methodsFor: 'event trace' stamp: 'ar 5/17/2002 20:18'!
eventTracer: aTracer
	^self setProperty: #eventTracer toValue: aTracer! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 5/24/2005 19:07'!
hand
	^((self valueOfProperty: #hand) valueOn: island) ifNil:[self island project primaryHand].! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 5/24/2005 19:17'!
hand: aHand
	^self setProperty: #hand toValue: (aHand ifNotNil:[island asFarRef: aHand])! !

!ScriptProcess methodsFor: 'properties' stamp: 'ar 7/20/2001 22:41'!
hasProperty: propName
	"Answer whether the receiver has the given property.  Deemed to have it only if I have a property dictionary entry for it and that entry is neither nil nor false"
	| prop |
	properties == nil ifTrue: [^ false].
	prop := properties at: propName ifAbsent: [nil].
	prop == nil ifTrue: [^ false].
	prop == false ifTrue: [^ false].
	^ true! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 7/9/2001 17:20'!
isActive
	^scheduler isScriptActive: self.! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 3/30/2007 18:26'!
isActiveScript
	^scheduler activeScript == self! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 10/18/2002 22:06'!
isDone
	^(flags bitAnd: FlagScriptDone) == FlagScriptDone! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 7/9/2001 17:15'!
isScheduled
	^myLink notNil! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 2/17/2002 19:12'!
isScript
	^true! !

!ScriptProcess methodsFor: 'converting' stamp: 'ar 4/24/2002 14:48'!
isScriptDescriptorIn: aScript
	^aScript runningScript == self! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 7/7/2001 15:30'!
isScriptingProcess
	^true! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 10/18/2002 22:08'!
isStarted
	^(flags bitAnd: FlagScriptStarted) == FlagScriptStarted! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 10/18/2002 22:10'!
isStopRequested
	^(flags bitAnd: FlagScriptStopRequest) == FlagScriptStopRequest! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 9/9/2002 03:12'!
isSystemProcess
	^false! !

!ScriptProcess methodsFor: 'testing' stamp: 'ar 10/10/2007 18:35'!
isTerminated
	"Answer whether the receiver is terminated. Also include regular finishing via #isDone"
	^self isDone or:[super isTerminated]! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/24/2001 20:43'!
message
	^message! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/24/2001 20:44'!
message: aMessageSend
	message := aMessageSend! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 4/24/2002 15:10'!
messageName
	^message ifNotNil:[message selector]! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 7/24/2001 20:03'!
myEventMap
	^myEventMap! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 7/24/2001 20:03'!
myEventMap: aDictionary
	myEventMap := aDictionary.! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 5/22/2005 16:23'!
name
	^name ifNil:[self processName]
! !

!ScriptProcess methodsFor: 'private' stamp: 'ar 6/15/2009 13:56'!
newScript
	suspendedContext := [self privateRunMsg.
						   self suspend] asContext.
	flags := 0.
	priority := Processor initialScriptPriority.
	scheduler := Processor activeProcess scheduler.! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 4/23/2002 19:52'!
nextEventHandlerFor: anObject selector: eventName
	^self nextEventHandlerFor: anObject selector: eventName startingAt: myHandler! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 4/23/2002 19:51'!
nextEventHandlerFor: anObject selector: eventName startingAt: observer
	"Find an return the enclosing event handler scope for the object/event pair. This method should be really fast; right now we're just searching the entire chain (which is bad) - we need something that's fast for the innermost event handler."
	| next |
	next := observer.
	[next == nil] whileFalse:[
		(next signaler == anObject and:[next selector == eventName]) ifTrue:[^next].
		next := next nextHandler.
	].
	^nil! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 4/23/2002 19:44'!
observe: eventSpec
	^self observe: eventSpec semaphore: nil buffer: nil! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 4/23/2002 19:44'!
observe: eventSpec semaphore: aSemaphore
	^self observe: eventSpec semaphore: aSemaphore buffer: nil! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 5/30/2007 15:20'!
observe: eventSpec semaphore: aSemaphore buffer: aBuffer
	"Install a observer for each event specified in the event spec.
	If a semaphore and buffer are provided, use those.
	This method will return the outermost handler (that is the one created
	for the first event) so that it's easy for the sender to release an entire
	chain of handlers."	
	| sema anObject eventName handler topHandler |
	topHandler := nil.
	sema := aSemaphore ifNil:[Semaphore new].
	1 to: eventSpec size // 2 do:[:i|
		anObject := eventSpec at: 2*i-1.
		eventName := eventSpec at: 2*i.

		handler := ScriptEventObserver new.
		topHandler ifNil:[topHandler := handler].
		handler semaphore: sema.
		handler buffer: aBuffer.
"@@@ do we :=really:= need the object/eventName pair here??? @@@"
		handler signaler: anObject.
		handler selector: eventName.
		handler resumeSuspending: 
			(self nextEventHandlerFor: anObject selector: eventName).
		handler nextHandler: myHandler.
		myHandler := handler.
		anObject on: eventName notify: handler].
	^topHandler! !

!ScriptProcess methodsFor: 'private-events' stamp: 'bf 7/12/2006 17:37'!
observe: eventSpec semaphore: aSemaphore buffer: aBuffer during: aBlock
	| handler |
	handler := self observe: eventSpec semaphore: aSemaphore buffer: aBuffer.
	^[aBlock value: handler] 
		ensure:[myHandler := myHandler releaseTo: handler nextHandler].! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 10/18/2002 22:07'!
on: eventName notify: anObserver
	eventName == #waiting ifTrue:[flags := flags bitOr: FlagScriptSignalWait].
	eventName == #resuming ifTrue:[flags := flags bitOr: FlagScriptSignalResume].
	^super on: eventName notify: anObserver! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/16/2001 23:08'!
owner
	^owner! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/16/2001 23:08'!
owner: anObject
	owner := anObject! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 2/18/2002 21:54'!
primitiveWaitUntil: eventName
	(eventName == #started and:[self isStarted]) ifTrue:[^self].
	(eventName == #done and:[self isDone]) ifTrue:[^self].
	^super primitiveWaitUntil: eventName.! !

!ScriptProcess methodsFor: 'printing' stamp: 'ar 4/24/2002 17:22'!
printOn: aStream
	aStream nextPutAll: self class name.
	aStream nextPutAll: ' in '.
	(self messageName notNil and:[self owner notNil]) ifTrue:[
		aStream print: self owner; nextPutAll:'>>'; nextPutAll: self messageName.
	] ifFalse:[
		suspendedContext printOn: aStream.
	].! !

!ScriptProcess methodsFor: 'private' stamp: 'ar 4/24/2002 14:26'!
privateRunMsg
	"Private. Evaluate the associated message."
	self startScriptProcess.
	result := message value. "if any"
	self stopScriptProcess: result.
! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 4/24/2002 16:36'!
processName
	"For the process browser"
	^String streamContents:[:s|
		s nextPutAll: (self hash asString forceTo: 5 paddingStartWith: $ ).
		s nextPutAll:' ['.
		s nextPutAll: (self owner receiver printString contractTo: 50).
		s nextPutAll:'>>'.
		s nextPutAll: (self messageName ifNil:['???']).
		s nextPut:$].
	].! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 4/23/2002 19:48'!
releaseObserver: aHandler
	myHandler := myHandler releaseTo: aHandler nextHandler.! !

!ScriptProcess methodsFor: 'private-events' stamp: 'ar 10/18/2002 22:08'!
removeEvent: eventName map: eventMap
	super removeEvent: eventName map: eventMap.
	(eventName == #waiting) 
		ifTrue:[flags := flags bitClear: FlagScriptSignalWait].
	(eventName == #resuming) 
		ifTrue:[flags := flags bitClear: FlagScriptSignalResume].! !

!ScriptProcess methodsFor: 'properties' stamp: 'ar 10/20/2002 15:44'!
removeProperty: propName
	| value |
	properties == nil ifTrue: [^ nil].
	value := properties removeKey: propName ifAbsent: [].
	properties size == 0 ifTrue: [properties := nil].
	^value! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 10/18/2002 22:09'!
requestStop
	"Request this script to be stopped"
	flags := flags bitOr: FlagScriptStopRequest.! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/7/2001 15:31'!
result
	self waitUntil: #done.
	^result! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/7/2001 15:31'!
result: aValue
	result := aValue! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 7/7/2001 16:32'!
scheduler
	^scheduler! !

!ScriptProcess methodsFor: 'properties' stamp: 'ar 7/20/2001 22:42'!
setProperty: propName toValue: aValue
	"These special cases move old properties into named fields of the extension"
	properties == nil ifTrue: [properties := IdentityDictionary new].
	properties at: propName put: aValue.
! !

!ScriptProcess methodsFor: 'initialize-release' stamp: 'ar 5/17/2002 20:00'!
setScheduler: aUserScriptScheduler
	scheduler := aUserScriptScheduler.
! !

!ScriptProcess methodsFor: 'initialize-release' stamp: 'ar 8/10/2009 17:17'!
startScriptProcess
	"Activate the receiver in the context of the (already provided) user script scheduler."
	scheduler ifNil:[self suspend].
	myLink := scheduler scheduleScript: self.
	"Note: From here on we are synchronized with the scheduler."
	flags := flags bitOr: FlagScriptStarted.
	self endWait.
	myEventMap ifNotNil:[self signal: #started].
! !

!ScriptProcess methodsFor: 'initialize-release' stamp: 'ar 8/10/2009 17:21'!
stopScriptProcess: resultObj
	"Finish the receiver in the context of the user script scheduler."
	myLink ifNotNil:[scheduler unscheduleScript: myLink].
	myLink := nil.
	self beginWait.
	result := resultObj.
	flags := flags bitOr: FlagScriptDone.
	owner ifNotNil:[owner scriptStopped: self].
	myHandler ifNotNil:[myHandler releaseTo: nil].
	myEventMap ifNotNil:[self signal: #done with: resultObj].
! !

!ScriptProcess methodsFor: 'initialize-release' stamp: 'ar 9/8/2009 14:58'!
terminate
	| list |
	self isDone ifTrue:[^self].
	self isActiveProcess ifFalse:[
		"avoid runaway"
		list := self suspend.
		list ifNotNil:[priority := priority - 1].
	].
	self stopScriptProcess: nil.
	super terminate.! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 11/12/2004 21:48'!
topContext
	^topContext! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 11/12/2004 21:48'!
topContext: aContext
	topContext := aContext! !

!ScriptProcess methodsFor: 'event trace' stamp: 'ar 5/17/2002 20:38'!
traceEvent: anEvent message: aMessage script: aScript
	^self eventTracer traceEvent: anEvent message: aMessage script: aScript from: self! !

!ScriptProcess methodsFor: 'event trace' stamp: 'ar 10/18/2002 22:08'!
traceEvents
	^(flags bitAnd: FlagScriptTraceEvents) == FlagScriptTraceEvents! !

!ScriptProcess methodsFor: 'event trace' stamp: 'ar 10/18/2002 22:07'!
traceEvents: aBool
	flags := aBool ifTrue:[flags bitOr: FlagScriptTraceEvents] ifFalse:[flags bitClear: FlagScriptTraceEvents].! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 2/13/2002 18:22'!
transferTo: aList
	"Transfer the receiver to another list to wait on"
	myList ifNil:[self error:'Not on any list'].
	myList remove: self.
	aList addLast: self.
	myList := aList.! !

!ScriptProcess methodsFor: 'properties' stamp: 'ar 7/20/2001 22:42'!
valueOfProperty: propName 
	^ self valueOfProperty: propName ifAbsent: [nil]! !

!ScriptProcess methodsFor: 'properties' stamp: 'ar 7/20/2001 22:42'!
valueOfProperty: propName ifAbsent: aBlock
	properties == nil ifTrue: [^ aBlock value].
	^ properties at: propName ifAbsent: [^ aBlock value]! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 5/22/2005 16:26'!
waitProcessOn: aWaitable
	"Wait on the given waitable object (Semaphore or Delay).
	Resynchronize with the scheduler when the wait completed."
	self == Processor activeProcess ifFalse:[^self error:'Not the active process'].
	self beginWait.
	aWaitable primitiveWait.
	"Note: It is not necessary to yield here - scripts are :=always:= synchronized so either the wait completed synchronuously in which case it will continue anyways. If not, then the following the scheduler wait will solve the problem for us."
	"Resynchronize the receiver with the scheduler."
	(self isActiveScript)
		ifFalse:[scheduler activeList primitiveWait].
	"Note: The reason for not resynchronizing when we're the active script is so that certain 'atomic' notions can be preserved. E.g., when we enter a signaled semaphore, the script should :=not:= be rescheduled at the end of the scheduler since the meaning of entering a signaled semaphore is non-preemptive."
	self endWait.
	^nil! !

!ScriptProcess methodsFor: 'synchronization' stamp: 'ar 7/7/2001 15:19'!
waitUntil: eventName
	(eventName == #started and:[self isStarted]) ifTrue:[^self].
	(eventName == #done and:[self isDone]) ifTrue:[^self].
	^super waitUntil: eventName.! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 2/7/2004 16:21'!
world
	^(self valueOfProperty: #world) ifNil:[super world]! !

!ScriptProcess methodsFor: 'accessing' stamp: 'ar 2/7/2004 16:22'!
world: aPlayer
	^self setProperty: #world toValue: aPlayer! !

!Heap methodsFor: '*Scripting-Extensions' stamp: 'ar 3/5/2002 01:29'!
firstOrNil
	tally = 0 ifTrue:[^nil] ifFalse:[^array at: 1]! !
AsyncScriptMessageSend initialize!
ProcessorScheduler initialize!
ScriptProcess initialize!
